# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""description of revision

Revision ID: 6185f1633a3d
Revises: b3adaf631bab
Create Date: 2017-02-19 00:00:00.000000

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '6185f1633a3d'
down_revision = 'b3adaf631bab'


def upgrade():
    op.add_column('sfc_portpair_details',
                  sa.Column('correlation',
                            sa.String(length=255),
                            nullable=True))



# Copyright 2015 Huawei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.common import ovs_lib
from neutron.plugins.ml2.drivers.openvswitch.agent.common import constants \
    as ovs_consts
from neutron_lib import exceptions
from oslo_log import log as logging

from networking_sfc._i18n import _

LOG = logging.getLogger(__name__)


def get_port_mask(min_port, max_port):
    """get port/mask serial by port range."""
    if min_port < 1 or max_port > 0xffff or min_port > max_port:
        msg = _("the port range is invalid")
        raise exceptions.InvalidInput(error_message=msg)
    masks = []
    while min_port <= max_port:
        mask = 0xffff
        while mask != 0:
            next_mask = (mask << 1) & 0xffff
            port_start = min_port & next_mask
            port_end = min_port + (next_mask ^ 0xffff)
            if port_start == min_port and port_end <= max_port:
                mask = next_mask
            else:
                break
        masks.append('0x%x/0x%x' % (min_port, mask))
        min_port = min_port + (mask ^ 0xffff) + 1

    return masks


class SfcOVSBridgeExt(object):

    def __init__(self, ovs_bridge):
        self.bridge = ovs_bridge

        # OpenFlow 1.3 is needed to manipulate groups
        # To support NSH feature, OpenFlow 1.3 is also needed
        self.bridge.use_at_least_protocol(ovs_consts.OPENFLOW13)

    # proxy most methods to self.bridge
    def __getattr__(self, name):
        return getattr(self.bridge, name)

    def do_action_groups(self, action, kwargs_list):
        group_strs = [_build_group_expr_str(kw, action) for kw in kwargs_list]
        if action == 'add' or action == 'del':
            cmd = '%s-groups' % action
        elif action == 'mod':
            cmd = '%s-group' % action
        else:
            msg = _("Action is illegal")
            raise exceptions.InvalidInput(error_message=msg)
        self.run_ofctl(cmd, ['-'], '\n'.join(group_strs))

    def add_group(self, **kwargs):
        self.do_action_groups('add', [kwargs])

    def mod_group(self, **kwargs):
        self.do_action_groups('mod', [kwargs])

    def delete_group(self, **kwargs):
        self.do_action_groups('del', [kwargs])

    def dump_group_for_id(self, group_id):
        retval = None
        group_str = "%d" % group_id
        group = self.run_ofctl("dump-groups", [group_str])
        if group:
            retval = '\n'.join(item for item in group.splitlines()
                               if ovs_lib.is_a_flow_line(item))
        return retval

    def get_bridge_ports(self):
        port_name_list = self.bridge.get_port_name_list()
        of_portno_list = list()
        for port_name in port_name_list:
            of_portno_list.append(self.bridge.get_port_ofport(port_name))
        return of_portno_list


def _build_group_expr_str(group_dict, cmd):
    group_expr_arr = []
    buckets = None
    group_id = None

    if cmd != 'del':
        if "group_id" not in group_dict:
            msg = _("Must specify one group Id on group addition"
                    " or modification")
            raise exceptions.InvalidInput(error_message=msg)
        group_id = "group_id=%s" % group_dict.pop('group_id')

        if "buckets" not in group_dict:
            msg = _("Must specify one or more buckets on group addition"
                    " or modification")
            raise exceptions.InvalidInput(error_message=msg)
        buckets = "%s" % group_dict.pop('buckets')

    if group_id:
        group_expr_arr.append(group_id)

    for key, value in group_dict.items():
        group_expr_arr.append("%s=%s" % (key, value))

    if buckets:
        group_expr_arr.append(buckets)

    return ','.join(group_expr_arr)



# Copyright (c) 2017 One Convergence Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

"""add tap_enabled attribute to port-pair-group

Revision ID: d6fb381b65f2
Revises: a3ad63aa834f
Create Date: 2017-08-03 13:57:59.908621

"""

# revision identifiers, used by Alembic.
revision = 'd6fb381b65f2'
down_revision = '53ed5bec6cff'


def upgrade():
    op.add_column('sfc_port_pair_groups',
                  sa.Column('tap_enabled',
                            sa.Boolean,
                            server_default=sa.sql.false(),
                            nullable=False)
                  )



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers

from networking_sfc.services.flowclassifier.common import exceptions as exc
from networking_sfc.services.flowclassifier.drivers import base as fc_driver


class OVSFlowClassifierDriver(fc_driver.FlowClassifierDriverBase):
    """FlowClassifier Driver Base Class."""

    def initialize(self):
        pass

    @log_helpers.log_method_call
    def create_flow_classifier(self, context):
        pass

    @log_helpers.log_method_call
    def update_flow_classifier(self, context):
        pass

    @log_helpers.log_method_call
    def delete_flow_classifier(self, context):
        pass

    @log_helpers.log_method_call
    def create_flow_classifier_precommit(self, context):
        """OVS Driver precommit before transaction committed.

        Make sure the logical_source_port is not None.
        """
        flow_classifier = context.current
        logical_source_port = flow_classifier['logical_source_port']
        if logical_source_port is None:
            raise exc.FlowClassifierBadRequest(message=(
                'FlowClassifier %s does not set '
                'logical source port in ovs driver' % flow_classifier['id']))



# Copyright 2015 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys
from unittest import mock

from neutronclient import shell
from neutronclient.tests.unit import test_cli20

from networking_sfc.cli import port_chain as pc

from oslo_utils import uuidutils

FAKE_port_pair_group1_UUID = uuidutils.generate_uuid()
FAKE_port_pair_group2_UUID = uuidutils.generate_uuid()
FAKE_FC1_UUID = uuidutils.generate_uuid()
FAKE_FC2_UUID = uuidutils.generate_uuid()
FAKE_PARAM1_UUID = uuidutils.generate_uuid()
FAKE_PARAM2_UUID = uuidutils.generate_uuid()


class CLITestV20PortChainExtensionJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        super(CLITestV20PortChainExtensionJSON, self).setUp()
        self._mock_extension_loading()
        self.register_non_admin_status_resource('port_chain')

    def _create_patch(self, name, func=None):
        patcher = mock.patch(name)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        port_chain = self._create_patch(ext_pkg +
                                        '._discover_via_entry_points')
        port_chain.return_value = [("port_chain", pc)]
        return port_chain

    def test_ext_cmd_loaded(self):
        neutron_shell = shell.NeutronShell('2.0')
        ext_cmd = {'port-chain-list': pc.PortChainList,
                   'port-chain-create': pc.PortChainCreate,
                   'port-chain-update': pc.PortChainUpdate,
                   'port-chain-delete': pc.PortChainDelete,
                   'port-chain-show': pc.PortChainShow}
        for cmd_name, cmd_class in ext_cmd.items():
            found = neutron_shell.command_manager.find_command([cmd_name])
            self.assertEqual(cmd_class, found[0])

    def test_create_port_chain_with_mandatory_param(self):
        """Create port_chain: myname."""
        resource = 'port_chain'
        cmd = pc.PortChainCreate(test_cli20.MyApp(sys.stdout),
                                 None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--port-pair-group', FAKE_port_pair_group1_UUID]
        position_names = ['name', 'port_pair_groups']
        position_values = [name, [FAKE_port_pair_group1_UUID]]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_chain_with_multiple_port_pair_group(self):
        """Create port_chain: myname."""
        resource = 'port_chain'
        cmd = pc.PortChainCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--port-pair-group', FAKE_port_pair_group1_UUID,
                '--port-pair-group', FAKE_port_pair_group2_UUID]
        position_names = ['name', 'port_pair_groups']
        position_values = [name, [FAKE_port_pair_group1_UUID,
                                  FAKE_port_pair_group2_UUID]]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_chain_with_all_params(self):
        """Create port_chain: myname."""
        resource = 'port_chain'
        cmd = pc.PortChainCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        desc = 'check port chain cli'
        chain_parameter = "correlation=mpls"
        chain_parameter_expected = {"correlation": "mpls"}
        args = [name, '--description', desc, '--port-pair-group',
                FAKE_port_pair_group1_UUID, '--flow-classifier',
                FAKE_FC1_UUID, '--chain-parameters', chain_parameter]
        position_names = ['name', 'description', 'port_pair_groups',
                          'flow_classifiers', 'chain_parameters']
        position_values = [name, desc, [FAKE_port_pair_group1_UUID],
                           [FAKE_FC1_UUID], chain_parameter_expected]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_chain_with_single_classifier(self):
        """Create port_chain: myname."""
        resource = 'port_chain'
        cmd = pc.PortChainCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--port-pair-group', FAKE_port_pair_group1_UUID,
                '--flow-classifier', FAKE_FC1_UUID]
        position_names = ['name', 'port_pair_groups', 'flow_classifiers']
        position_values = [name, [FAKE_port_pair_group1_UUID], [FAKE_FC1_UUID]]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_chain_with_chain_parameters(self):
        """Create port_chain: myname."""
        resource = 'port_chain'
        cmd = pc.PortChainCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--port-pair-group', FAKE_port_pair_group1_UUID,
                '--chain-parameters', 'symmetric=True']
        position_names = ['name', 'port_pair_groups', 'chain_parameters']
        position_values = [name, [FAKE_port_pair_group1_UUID],
                           {'symmetric': 'True'}]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_port_chain_with_multiple_classifiers(self):
        """Create port_chain: myname."""
        resource = 'port_chain'
        cmd = pc.PortChainCreate(test_cli20.MyApp(sys.stdout), None)
        name = 'myname'
        myid = 'myid'
        args = [name, '--port-pair-group', FAKE_port_pair_group1_UUID,
                '--flow-classifier', FAKE_FC1_UUID,
                '--flow-classifier', FAKE_FC2_UUID]
        position_names = ['name', 'port_pair_groups', 'flow_classifiers']
        position_values = [name, [FAKE_port_pair_group1_UUID], [FAKE_FC1_UUID,
                                                                FAKE_FC2_UUID]]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_update_port_chain(self):
        """Update port_chain: myid --name myname."""
        resource = 'port_chain'
        cmd = pc.PortChainUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname'],
                                   {'name': 'myname'})

    def test_update_port_chain_with_no_flow_classifier(self):
        """Update port_chain: myid --name myname --no-flow-classifier None."""
        resource = 'port_chain'
        cmd = pc.PortChainUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname',
                                    '--no-flow-classifier'],
                                   {'name': 'myname',
                                    'flow_classifiers': []})

    def test_update_port_chain_with_single_port_pair_group(self):
        """Update port_chain: myid --name myname --port-pair-group uuid."""
        resource = 'port_chain'
        cmd = pc.PortChainUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname',
                                    '--port-pair-group',
                                    FAKE_port_pair_group1_UUID,
                                    '--no-flow-classifier'],
                                   {'name': 'myname',
                                    'port_pair_groups': [
                                        FAKE_port_pair_group1_UUID],
                                    'flow_classifiers': []})

    def test_update_port_chain_with_multi_port_pair_groups(self):
        """Update port_chain: myid --name myname --port-pair-group uuid ..."""
        resource = 'port_chain'
        cmd = pc.PortChainUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname',
                                    '--port-pair-group',
                                    FAKE_port_pair_group1_UUID,
                                    '--port-pair-group',
                                    FAKE_port_pair_group2_UUID,
                                    '--no-flow-classifier'],
                                   {'name': 'myname',
                                    'port_pair_groups': [
                                        FAKE_port_pair_group1_UUID,
                                        FAKE_port_pair_group2_UUID],
                                    'flow_classifiers': []})

    def test_update_port_chain_with_single_classifier(self):
        """Update port_chain: myid --name myname --flow-classifier uuid."""
        resource = 'port_chain'
        cmd = pc.PortChainUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname',
                                    '--flow-classifier', FAKE_FC1_UUID],
                                   {'name': 'myname',
                                    'flow_classifiers': [FAKE_FC1_UUID]})

    def test_update_port_chain_with_multi_classifiers(self):
        """Update port_chain: myid --name myname --flow-classifier uuid ..."""
        resource = 'port_chain'
        cmd = pc.PortChainUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname',
                                    '--flow-classifier', FAKE_FC1_UUID,
                                    '--flow-classifier', FAKE_FC2_UUID],
                                   {'name': 'myname',
                                    'flow_classifiers': [
                                        FAKE_FC1_UUID, FAKE_FC2_UUID]})

    def test_update_port_chain_with_port_pair_group_classifier(self):
        """Update port_chain."""
        resource = 'port_chain'
        cmd = pc.PortChainUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname',
                                    '--flow-classifier', FAKE_FC1_UUID,
                                    '--port-pair-group',
                                    FAKE_port_pair_group1_UUID],
                                   {'name': 'myname',
                                    'flow_classifiers': [FAKE_FC1_UUID],
                                    'port_pair_groups': [
                                        FAKE_port_pair_group1_UUID]})

    def test_delete_port_chain(self):
        """Delete port-chain: myid."""
        resource = 'port_chain'
        cmd = pc.PortChainDelete(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid]
        self._test_delete_resource(resource, cmd, myid, args)

    def test_list_port_chain(self):
        """List port_chain."""
        resources = 'port_chains'
        cmd = pc.PortChainList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_list_port_chains_sort(self):
        """List port_chains: --sort-key name --sort-key id --sort-key asc

        --sort-key desc
        """
        resources = "port_chains"
        cmd = pc.PortChainList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"])

    def test_show_port_chain(self):
        """Show port-chain: --fields id --fields name myid."""
        resource = 'port_chain'
        cmd = pc.PortChainShow(test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id,
                                 args, ['id', 'name'])



# Copyright 2017 Futurewei. All rights reserved.
# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import models_v2
import neutron.plugins.ml2.drivers.l2pop.db as l2pop_db
import neutron.plugins.ml2.drivers.l2pop.rpc as l2pop_rpc
from neutron_lib import constants as const
from neutron_lib import context as n_context
from neutron_lib.db import model_query
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_serialization import jsonutils

from networking_sfc.extensions import flowclassifier
from networking_sfc.extensions import sfc
from networking_sfc.services.sfc.common import exceptions as exc
from networking_sfc.services.sfc.drivers import base as driver_base
from networking_sfc.services.sfc.drivers.ovs import constants as ovs_const
from networking_sfc.services.sfc.drivers.ovs import db as ovs_sfc_db
from networking_sfc.services.sfc.drivers.ovs import rpc as ovs_sfc_rpc
from networking_sfc.services.sfc.drivers.ovs import rpc_topics as sfc_topics

LOG = logging.getLogger(__name__)


class OVSSfcDriver(driver_base.SfcDriverBase,
                   ovs_sfc_db.OVSSfcDriverDB):
    """Sfc Driver Base Class."""

    def initialize(self):
        super(OVSSfcDriver, self).initialize()
        self.ovs_driver_rpc = ovs_sfc_rpc.SfcAgentRpcClient(
            sfc_topics.SFC_AGENT
        )
        self.rpc_ctx = n_context.get_admin_context_without_session()
        self._setup_rpc()

    def _setup_rpc(self):
        # Setup a rpc server
        self.topic = sfc_topics.SFC_PLUGIN
        self.endpoints = [ovs_sfc_rpc.SfcRpcCallback(self)]
        self.conn = n_rpc.Connection()
        self.conn.create_consumer(self.topic, self.endpoints, fanout=False)
        self.conn.consume_in_threads()

    def _get_port_infos(self, context, port, segment, agent_host):
        if not agent_host:
            return

        agent = l2pop_db.get_agent_by_host(context, agent_host)
        if not agent:
            return

        agent_ip = l2pop_db.get_agent_ip(agent)
        if not agent_ip:
            LOG.warning("Unable to retrieve the agent ip, check the agent "
                        "configuration.")
            return

        if not segment:
            LOG.warning("Port %(port)s updated by agent %(agent)s "
                        "isn't bound to any segment",
                        {'port': port['id'], 'agent': agent})
            return

        network_types = l2pop_db.get_agent_l2pop_network_types(agent)
        if network_types is None:
            network_types = l2pop_db.get_agent_tunnel_types(agent)
        if segment['network_type'] not in network_types:
            return

        fdb_entries = [l2pop_rpc.PortInfo(mac_address=port['mac_address'],
                                          ip_address=ip['ip_address'])
                       for ip in port['fixed_ips']]
        return agent_ip, fdb_entries

    @log_helpers.log_method_call
    def _get_agent_fdb(self, context, port, segment, agent_host):
        agent_ip, port_fdb_entries = self._get_port_infos(context,
                                                          port,
                                                          segment,
                                                          agent_host)
        if not port_fdb_entries:
            return

        network_id = port['network_id']
        other_fdb_entries = {network_id:
                             {'segment_id': segment['segmentation_id'],
                              'network_type': segment['network_type'],
                              'ports': {agent_ip: []}}}

        # Agent is removing its last activated port in this network,
        # other agents needs to be notified to delete their flooding entry.
        other_fdb_entries[network_id]['ports'][agent_ip].append(
            const.FLOODING_ENTRY)
        # Notify other agents to remove fdb rules for current port
        if port['device_owner'] != const.DEVICE_OWNER_DVR_INTERFACE:
            fdb_entries = port_fdb_entries
            other_fdb_entries[network_id]['ports'][agent_ip] += fdb_entries

        return other_fdb_entries

    @log_helpers.log_method_call
    def _get_remote_pop_ports(self, context, flow_rule):
        pop_ports = []
        if not flow_rule.get('next_hops', None):
            return pop_ports
        pop_host = flow_rule['host_id']
        core_plugin = directory.get_plugin()
        drivers = core_plugin.mechanism_manager.mech_drivers
        l2pop_driver = drivers.get('l2population', None)
        if l2pop_driver is None:
            return pop_ports
        for next_hop in flow_rule['next_hops']:
            agent_active_ports = \
                l2pop_db.get_agent_network_active_port_count(
                    self.admin_context,
                    pop_host,
                    next_hop['net_uuid'])
            segment = {}
            if agent_active_ports == 0:
                filters = {'network_id': [next_hop['net_uuid']],
                           'mac_address': [next_hop['mac_address']]}
                ports = core_plugin.get_ports(self.admin_context,
                                              filters=filters)
                if not ports:
                    continue
                segment['network_type'] = next_hop['network_type']
                segment['segmentation_id'] = next_hop['segment_id']
                pop_ports.append((ports[0], segment))

        return pop_ports

    @log_helpers.log_method_call
    def _get_network_other_active_entry_count(self, host, remote_port_id):
        agent_active_ports = 0
        port_detail = self.get_port_detail_by_filter(
            dict(ingress=remote_port_id))
        for assoc in port_detail['path_nodes']:
            node = self.get_path_node(assoc['pathnode_id'])
            if node['node_type'] != ovs_const.SRC_NODE:
                filter = dict(nsp=node['nsp'],
                              nsi=node['nsi'] + 1)
                pre_node = self.get_path_node_by_filter(filter)
                if not pre_node:
                    continue
                for each in pre_node['portpair_details']:
                    pre_port = self.get_port_detail_by_filter(dict(id=each))
                    if host == pre_port['host_id']:
                        agent_active_ports += 1

        return agent_active_ports

    def _call_on_l2pop_driver(self, context, method_name, flow_rule):
        pop_host = flow_rule['host_id']
        pop_ports = self._get_remote_pop_ports(context, flow_rule)
        for (port, segment) in pop_ports:
            port_id = port['id']
            host_id = port['binding:host_id']
            active_entry_count = self._get_network_other_active_entry_count(
                pop_host,
                port_id)

            if active_entry_count == 1:
                fdb_entry = self._get_agent_fdb(
                    context,
                    port,
                    segment,
                    host_id)

                getattr(l2pop_rpc.L2populationAgentNotifyAPI(), method_name)(
                    self.rpc_ctx, fdb_entry, pop_host)

    def _update_agent_fdb_entries(self, context, flow_rule):
        self._call_on_l2pop_driver(context, "add_fdb_entries", flow_rule)

    def _delete_agent_fdb_entries(self, context, flow_rule):
        self._call_on_l2pop_driver(context, "remove_fdb_entries", flow_rule)

    def _get_subnet_by_port(self, id):
        core_plugin = directory.get_plugin()
        port = core_plugin.get_port(self.admin_context, id)
        subnet = None
        for ip in port['fixed_ips']:
            subnet = core_plugin.get_subnet(self.admin_context,
                                            ip["subnet_id"])
            # currently only support one subnet for a port
            break

        return subnet

    @log_helpers.log_method_call
    def _get_portgroup_members(self, context, pg_id, fwd_path):
        next_group_members = []
        ppg_obj = context._plugin._get_port_pair_group(context._plugin_context,
                                                       pg_id)
        group_intid = ppg_obj['group_id']
        tap_enabled = ppg_obj['tap_enabled']
        LOG.debug('group_intid: %s', group_intid)
        pg = context._plugin.get_port_pair_group(context._plugin_context,
                                                 pg_id)
        for pp_id in pg['port_pairs']:
            pp = context._plugin.get_port_pair(context._plugin_context, pp_id)
            filters = {}
            if pp.get('ingress', None):
                filters.update({'ingress': pp['ingress']})
            if pp.get('egress', None):
                filters.update({'egress': pp['egress']})
            pd = self.get_port_detail_by_filter(filters)
            if pd:
                next_group_members.append(
                    dict(portpair_id=pd['id'], weight=1,
                         tap_enabled=tap_enabled))
        if fwd_path is False:
            next_group_members.reverse()
        return group_intid, next_group_members

    def _get_port_pair_detail_by_port_pair(self, context, port_pair_id):
        pp = context._plugin.get_port_pair(context._plugin_context,
                                           port_pair_id)
        filters = {}
        if pp.get('ingress', None):
            filters.update({'ingress': pp['ingress']})
        if pp.get('egress', None):
            filters.update({'egress': pp['egress']})
        pd = self.get_port_detail_by_filter(filters)

        return pd

    @log_helpers.log_method_call
    def _add_flowclassifier_port_assoc(self, fc_ids, project_id,
                                       src_node):
        for fc in self._get_fcs_by_ids(fc_ids):
            need_assoc = True
            # lookup the source port, when it's reverse path
            # set logical_destination_port to be source port
            if src_node['fwd_path'] is False:
                src_pd_filter = dict(
                    egress=fc['logical_destination_port'],
                    project_id=project_id,
                )
            else:
                src_pd_filter = dict(
                    egress=fc['logical_source_port'],
                    project_id=project_id,
                )
            src_pd = self.get_port_detail_by_filter(src_pd_filter)

            if not src_pd:
                # Create source port detail
                src_pd = self._create_port_pair_detail(src_pd_filter)
                LOG.debug('create src port detail: %s', src_pd)
            else:
                for path_node in src_pd['path_nodes']:
                    if path_node['pathnode_id'] == src_node['id']:
                        need_assoc = False
            if need_assoc:
                # Create associate relationship
                assco_args = {
                    'portpair_id': src_pd['id'],
                    'pathnode_id': src_node['id'],
                    'weight': 1,
                }
                sna = self.create_pathport_assoc(assco_args)
                LOG.debug('create assoc src port with node: %s', sna)
                src_node['portpair_details'].append(src_pd['id'])

    def _remove_flowclassifier_port_assoc(self, fc_ids, project_id,
                                          src_nodes):
        if not fc_ids:
            return
        for src_node in src_nodes:
            for fc in self._get_fcs_by_ids(fc_ids):
                # delete source port detail
                if src_node['fwd_path'] is False:
                    src_pd_filter = dict(
                        egress=fc['logical_destination_port'],
                        project_id=project_id
                    )
                elif src_node['fwd_path']:
                    src_pd_filter = dict(
                        egress=fc['logical_source_port'],
                        project_id=project_id
                    )
                pds = self.get_port_details_by_filter(src_pd_filter)
                if pds:
                    for pd in pds:
                        # update src_node portpair_details refence info
                        if src_node and pd['id'] in src_node[
                            'portpair_details'
                        ]:
                            self.delete_pathport_assoc(src_node['id'],
                                                       pd['id'])
                            src_node['portpair_details'].remove(pd['id'])
                            # path_nodes is [] when passing from path_node
                            # not delete any src_node in portpair_details table
                            # why need to check len(pd['path_nodes']
                            # if len(pd['path_nodes']) == 1:
                            self.delete_port_pair_detail(pd['id'])

    @log_helpers.log_method_call
    def _create_portchain_path(self, context, port_chain, fwd_path):
        path_nodes = []
        # Create an assoc object for chain_id and path_id
        # context = context._plugin_context
        path_id = port_chain['chain_id']

        if not path_id:
            LOG.error('No path_id available for creating port chain path')
            return

        port_pair_groups = port_chain['port_pair_groups']
        sf_path_length = len(port_pair_groups)

        # Detect cross-subnet transit, returns ppg details
        ppg_details_list = self._validate_chain(context,
                                                port_chain,
                                                port_pair_groups,
                                                fwd_path)

        next_group_intid = None
        next_group_members = None
        previous_node_id = None
        # get the init and last port_pair_group
        if fwd_path:
            next_group_intid, next_group_members = self._get_portgroup_members(
                context, port_chain['port_pair_groups'][0], fwd_path)

        else:
            next_group_intid, next_group_members = self._get_portgroup_members(
                context,
                port_chain['port_pair_groups'][sf_path_length - 1],
                fwd_path)

        # Create a head node object for port chain
        src_args = {'project_id': port_chain['project_id'],
                    'node_type': ovs_const.SRC_NODE,
                    'nsp': path_id,
                    'nsi': 0xff,
                    'portchain_id': port_chain['id'],
                    'status': ovs_const.STATUS_BUILDING,
                    'next_group_id': next_group_intid,
                    'next_hop': jsonutils.dumps(next_group_members),
                    'fwd_path': fwd_path,
                    'ppg_n_tuple_mapping': None,
                    'tap_enabled': False
                    }
        src_node = self.create_path_node(src_args)
        LOG.debug('create src node: %s', src_node)
        path_nodes.append(src_node)
        previous_node_id = src_node['id']

        # Create a destination node object for port chain
        dst_args = {
            'project_id': port_chain['project_id'],
            'node_type': ovs_const.DST_NODE,
            'nsp': path_id,
            'nsi': 0xff - sf_path_length - 1,
            'portchain_id': port_chain['id'],
            'status': ovs_const.STATUS_BUILDING,
            'next_group_id': None,
            'next_hop': None,
            'fwd_path': fwd_path,
            'ppg_n_tuple_mapping': None,
            'tap_enabled': False
        }
        dst_node = self.create_path_node(dst_args)
        LOG.debug('create dst node: %s', dst_node)

        # need to pass project_id here
        self._add_flowclassifier_port_assoc(
            port_chain['flow_classifiers'],
            port_chain['project_id'],
            src_node
        )

        curr_group = context._plugin.get_port_pair_group(
            context._plugin_context, port_pair_groups[0])
        for i in range(sf_path_length):
            cur_group_members = next_group_members
            # next_group for next hop
            if i < sf_path_length - 1:
                if fwd_path:
                    next_group_intid, next_group_members = (
                        self._get_portgroup_members(
                            context, port_pair_groups[i + 1], fwd_path)
                    )
                elif fwd_path is False:
                    next_group_intid, next_group_members = (
                        self._get_portgroup_members(
                            context,
                            port_pair_groups[sf_path_length - 2 - i],
                            fwd_path)
                    )
            else:
                next_group_intid = None
                next_group_members = None

            # Get current port_pair_group based on current port_pair_group id
            if i < sf_path_length:
                if fwd_path:
                    curr_group = context._plugin.get_port_pair_group(
                        context._plugin_context, port_pair_groups[i])
                elif fwd_path is False:
                    curr_group = context._plugin.get_port_pair_group(
                        context._plugin_context,
                        port_pair_groups[sf_path_length - 1 - i])

            # Set curr_ppg_flag = 1, when current port_pair_group has
            # ppg_n_tuple_mapping dict in port_pair_group_parameters
            ppg_n_tuple_mapping = curr_group.get(
                'port_pair_group_parameters', None)
            if ppg_n_tuple_mapping:
                ppg_n_tuple_mapping = ppg_n_tuple_mapping.get(
                    'ppg_n_tuple_mapping', None)
                if ppg_n_tuple_mapping:
                    if ppg_n_tuple_mapping.get('ingress_n_tuple', None) or \
                            ppg_n_tuple_mapping.get('egress_n_tuple', None):
                        ppg_n_tuple_mapping['curr_ppg_flag'] = 1
                    else:
                        ppg_n_tuple_mapping = None

            # Create a node object
            node_args = {
                'project_id': port_chain['project_id'],
                'node_type': ovs_const.SF_NODE,
                'nsp': path_id,
                'nsi': 0xfe - i,
                'portchain_id': port_chain['id'],
                'status': ovs_const.STATUS_BUILDING,
                'next_group_id': next_group_intid,
                'next_hop': (
                    None if not next_group_members else
                    jsonutils.dumps(next_group_members)
                ),
                'fwd_path': fwd_path,
                'ppg_n_tuple_mapping': (
                    None if not ppg_n_tuple_mapping else
                    jsonutils.dumps(ppg_n_tuple_mapping)),
                'tap_enabled': ppg_details_list[i]['tap_enabled'],
                'previous_node_id': previous_node_id
            }
            sf_node = self.create_path_node(node_args)
            LOG.debug('chain path node: %s', sf_node)
            previous_node_id = sf_node['id']

            # Create the assocation objects that combine the pathnode_id with
            # the ingress of the port_pairs in the current group
            # when port_group does not reach tail
            for member in cur_group_members:
                assco_args = {'portpair_id': member['portpair_id'],
                              'pathnode_id': sf_node['id'],
                              'weight': member['weight'], }
                sfna = self.create_pathport_assoc(assco_args)
                LOG.debug('create assoc port with node: %s', sfna)
                sf_node['portpair_details'].append(member['portpair_id'])
            path_nodes.append(sf_node)

            path_nodes = self._update_ppg_n_tuple_in_flow_rule(
                path_nodes, fwd_path, sf_path_length)

        dst_node = self.update_path_node(
            dst_node['id'],
            {'previous_node_id': previous_node_id})
        path_nodes.append(dst_node)
        path_nodes = self._update_sc_path_parameter(path_nodes)
        return path_nodes

    # Function to update adjacent node ppg_n_tuple_mapping values if have
    # ppg_n_tuple_mapping in path_nodes
    def _update_ppg_n_tuple_in_flow_rule(self, path_nodes,
                                         fwd_path, sf_path_length):
        for index, node in enumerate(path_nodes):
            if not node['ppg_n_tuple_mapping']:
                # Update reverse SRC_NODE ppg_n_tuple_mapping based on last
                # forward SF_NODE ppg_n_tuple_mapping
                if (
                    node['node_type'] == ovs_const.SRC_NODE and
                    fwd_path is False
                ):
                    last_fwd_sf_node = self.get_path_node_by_filter(
                        filters={'portchain_id': node['portchain_id'],
                                 'nsi': 0xff - sf_path_length,
                                 'fwd_path': True}
                    )
                    if last_fwd_sf_node:
                        if last_fwd_sf_node['ppg_n_tuple_mapping']:
                            ppg_n_tuple_mapping = jsonutils.loads(
                                last_fwd_sf_node['ppg_n_tuple_mapping'])
                            if (
                                ppg_n_tuple_mapping.get(
                                    'ingress_n_tuple', None) or
                                ppg_n_tuple_mapping.get('egress_n_tuple', None)
                            ):
                                # Set curr_ppg_flag = 2 when
                                # ppg_n_tuple_mapping inherits from
                                # last_fwd_sf_node
                                ppg_n_tuple_mapping['curr_ppg_flag'] = 2
                                node = self.update_path_node(
                                    node['id'],
                                    {'ppg_n_tuple_mapping':
                                     None if not ppg_n_tuple_mapping else
                                     jsonutils.dumps(ppg_n_tuple_mapping)}
                                )
                                path_nodes[index] = node
                # Update SF_NODE ppg_n_tuple_mapping based on current
                # ppg_n_tuple_mapping and prev_node ppg_n_tuple_mapping
                elif node['node_type'] == ovs_const.SF_NODE:
                    if node['nsi'] == 0xfe:
                        prev_node = path_nodes[index - 2]
                    else:
                        prev_node = path_nodes[index - 1]
                    if prev_node:
                        if prev_node['ppg_n_tuple_mapping']:
                            ppg_n_tuple_mapping = jsonutils.loads(
                                prev_node['ppg_n_tuple_mapping'])
                            # Set curr_ppg_flag = 2 when
                            # ppg_n_tuple_mapping inherits from previous
                            # sf_node. Set curr_ppg_flag = 3 when
                            # ppg_n_tuple_mapping inherits from previous
                            # sf_node, and fwd_path is False
                            if (
                                ppg_n_tuple_mapping.get(
                                    'ingress_n_tuple', None) or
                                ppg_n_tuple_mapping.get('egress_n_tuple', None)
                            ):
                                if (
                                    ppg_n_tuple_mapping[
                                        'curr_ppg_flag'] == 1 and
                                    fwd_path is False
                                ):
                                    ppg_n_tuple_mapping['curr_ppg_flag'] = 3
                                else:
                                    ppg_n_tuple_mapping['curr_ppg_flag'] = 2
                                node = self.update_path_node(
                                    node['id'],
                                    {'ppg_n_tuple_mapping':
                                     None if not ppg_n_tuple_mapping else
                                     jsonutils.dumps(ppg_n_tuple_mapping)}
                                )
                                path_nodes[index] = node
        return path_nodes

    def _delete_path_node_port_flowrule(self, context, node,
                                        port, pc_corr, fc_ids):
        # if this port is not binding, don't to generate flow rule
        if not port['host_id']:
            return
        flow_rule = self._build_portchain_flowrule_body(node,
                                                        port,
                                                        pc_corr,
                                                        del_fc_ids=fc_ids)
        self.ovs_driver_rpc.ask_agent_to_delete_flow_rules(self.admin_context,
                                                           flow_rule)
        self._delete_agent_fdb_entries(context, flow_rule)

    def _delete_path_node_flowrule(self, context, node, pc_corr, fc_ids):
        if node['portpair_details'] is None:
            return
        for each in node['portpair_details']:
            port = self.get_port_detail_by_filter(dict(id=each))
            if port:
                self._delete_path_node_port_flowrule(
                    context, node, port, pc_corr, fc_ids)

    @log_helpers.log_method_call
    def _delete_portchain_path(self, context, port_chain):
        pds = self.get_path_nodes_by_filter(
            dict(portchain_id=port_chain['id']))
        src_nodes = []
        if pds:
            for pd in pds:
                if pd['node_type'] == ovs_const.SRC_NODE:
                    src_nodes.append(pd)
                pc_corr = port_chain['chain_parameters']['correlation']
                updated_pd = self._update_tap_enabled_node(pd)
                self._delete_path_node_flowrule(
                    context,
                    updated_pd,
                    pc_corr,
                    port_chain['flow_classifiers']
                )
                self._build_tap_ingress_flow(pc_corr,
                                             updated_pd,
                                             pd['fwd_path'],
                                             delete=True)
            for pd in pds:
                self.delete_path_node(pd['id'])

        # delete the ports on the traffic classifier
        self._remove_flowclassifier_port_assoc(
            port_chain['flow_classifiers'],
            port_chain['project_id'],
            src_nodes
        )

    def _update_path_node_next_hops(self, flow_rule):
        node_next_hops = []
        if not flow_rule['next_hop']:
            return None
        next_hops = jsonutils.loads(flow_rule['next_hop'])
        if not next_hops:
            return None
        core_plugin = directory.get_plugin()
        for member in next_hops:
            detail = {}
            port_detail = self.get_port_detail_by_filter(
                dict(id=member['portpair_id']))
            if not port_detail or not port_detail['host_id']:
                continue
            detail['local_endpoint'] = port_detail['local_endpoint']
            detail['weight'] = member['weight']
            detail['mac_address'] = port_detail['mac_address']
            detail['in_mac_address'] = port_detail['in_mac_address']
            detail['segment_id'] = port_detail['segment_id']
            detail['network_type'] = port_detail['network_type']
            detail['pp_corr'] = port_detail['correlation']
            port = core_plugin.get_port(
                self.admin_context, port_detail['ingress'])
            detail['net_uuid'] = port['network_id']
            detail['tap_enabled'] = member['tap_enabled']
            if member['tap_enabled']:
                self._update_next_hop_details(flow_rule, port_detail, detail)
            node_next_hops.append(detail)
        flow_rule['next_hops'] = node_next_hops
        flow_rule.pop('next_hop')

        return node_next_hops

    # As of the "no-SFC-proxy" MPLS correlation support, pc_corr is passed.
    # pc_corr is expected to be the port-chain's correlation parameter, i.e.
    # the chain-wide SFC Encapsulation protocol. This is necessary to compare
    # with port-pairs' correlations and decide whether SFC Proxy is needed.
    def _build_portchain_flowrule_body(self, node, port, pc_corr,
                                       add_fc_ids=None, del_fc_ids=None):
        node_info = node.copy()
        node_info.pop('project_id')
        node_info.pop('portpair_details')

        port_info = port.copy()
        port_info.pop('project_id')
        port_info.pop('id')
        port_info.pop('path_nodes')
        # port_info.pop('host_id')

        # change egress port in src_nodes
        if(
            node_info['fwd_path'] is False and node_info['node_type'] ==
            ovs_const.SRC_NODE
        ):
            if add_fc_ids is not None:
                fcs = self._get_fcs_by_ids(add_fc_ids)
            elif del_fc_ids is not None:
                fcs = self._get_fcs_by_ids(del_fc_ids)
            for fc in fcs:
                if fc['logical_source_port'] == port_info['egress']:
                    port_info['egress'] = fc['logical_destination_port']

        flow_rule = dict(node_info, **port_info)
        flow_rule['pc_corr'] = pc_corr
        if node_info['node_type'] != ovs_const.SRC_NODE:
            flow_rule['pp_corr'] = port_info.get('correlation')
        else:
            # there's no correlation for src nodes
            flow_rule['pp_corr'] = None
        flow_rule.pop('correlation')  # correlation becomes simply pp_corr

        # if this port belongs to an SFC Encapsulation-aware VM,
        # only notify the flow classifier for the 1st SF.
        flow_rule['add_fcs'] = self._filter_flow_classifiers(
            flow_rule, add_fc_ids)
        flow_rule['del_fcs'] = self._filter_flow_classifiers(
            flow_rule, del_fc_ids)

        self._update_portchain_group_reference_count(flow_rule,
                                                     port['host_id'])
        # update next hop info
        self._update_path_node_next_hops(flow_rule)

        return flow_rule

    def _filter_flow_classifiers(self, flow_rule, fc_ids):
        """Filter flow classifiers.

        @return: list of the flow classifiers
        """

        fc_return = []

        if not fc_ids:
            return fc_return
        fcs = self._get_fcs_by_ids(fc_ids)
        for fc in fcs:
            new_fc = fc.copy()
            new_fc.pop('id')
            new_fc.pop('name')
            new_fc.pop('project_id')
            new_fc.pop('description')
            router_ints = const.ROUTER_INTERFACE_OWNERS
            logical_source_port = new_fc['logical_source_port']
            if logical_source_port is not None:
                port_src = model_query.get_by_id(
                    self.admin_context, models_v2.Port, logical_source_port
                )
                if (
                    new_fc['source_ip_prefix'] is None and
                    port_src['device_owner'] not in router_ints
                ):
                    src_ips = port_src['fixed_ips']
                    # For now, only handle when the port has a single IP
                    if len(src_ips) == 1:
                        new_fc['source_ip_prefix'] = src_ips[0]['ip_address']

            logical_destination_port = new_fc['logical_destination_port']
            if logical_destination_port is not None:
                port_dst = model_query.get_by_id(
                    self.admin_context, models_v2.Port,
                    logical_destination_port
                )
                if (
                    new_fc['destination_ip_prefix'] is None and
                    port_dst['device_owner'] not in router_ints
                ):
                    dst_ips = port_dst['fixed_ips']
                    # For now, only handle when the port has a single IP
                    if len(dst_ips) == 1:
                        new_fc['destination_ip_prefix'] = (
                            dst_ips[0]['ip_address']
                        )
            # Update new_fc n tuple info based flow_rule['ppg_n_tuple_mapping']
            # and flow_rule['fwd_path']
            if flow_rule['ppg_n_tuple_mapping']:
                ppg_n_tuple_mapping = jsonutils.loads(
                    flow_rule['ppg_n_tuple_mapping'])
                if (
                    flow_rule['fwd_path'] is False and
                    ppg_n_tuple_mapping['curr_ppg_flag'] == 1 or
                    ppg_n_tuple_mapping['curr_ppg_flag'] == 3
                ):
                    for ingress_key, ingress_value in \
                            ppg_n_tuple_mapping['ingress_n_tuple'].items():
                        new_fc[ingress_key] = ingress_value
                else:
                    for egress_key, egress_value in \
                            ppg_n_tuple_mapping['egress_n_tuple'].items():
                        new_fc[egress_key] = egress_value

            if flow_rule['fwd_path'] is False:
                # swap logical_source_port & logical_destination_port
                new_fc['logical_source_port'] = fc['logical_destination_port']
                new_fc['logical_destination_port'] = fc['logical_source_port']

            if (
                # add_flow & del_flow in flowrule pass into agent
                flow_rule['node_type'] == ovs_const.SRC_NODE and
                flow_rule['egress'] == new_fc['logical_source_port']
            ):
                fc_return.append(new_fc)

            elif flow_rule['node_type'] == ovs_const.SF_NODE:
                fc_return.append(new_fc)

        return fc_return

    def _update_path_node_port_flowrules(self, context, node, port, pc_corr,
                                         add_fc_ids=None, del_fc_ids=None):
        # if this port is not binding, don't to generate flow rule
        if not port['host_id']:
            return
        flow_rule = self._build_portchain_flowrule_body(node,
                                                        port,
                                                        pc_corr,
                                                        add_fc_ids=add_fc_ids,
                                                        del_fc_ids=del_fc_ids)
        self.ovs_driver_rpc.ask_agent_to_update_flow_rules(self.admin_context,
                                                           flow_rule)
        self._update_agent_fdb_entries(context, flow_rule)

    def _update_path_node_flowrules(self, context, node, pc_corr,
                                    add_fc_ids=None, del_fc_ids=None):
        if node['portpair_details'] is None:
            return
        updated_node = self._update_tap_enabled_node(node)
        for each in updated_node['portpair_details']:
            port = self.get_port_detail_by_filter(dict(id=each))
            if port:
                self._update_path_node_port_flowrules(
                    context, updated_node,
                    port, pc_corr, add_fc_ids, del_fc_ids)
        self._build_tap_ingress_flow(pc_corr, updated_node, node['fwd_path'])

    def _update_path_nodes(self, context, nodes, pc_corr,
                           add_fc_ids=None, del_fc_ids=None):
        for node in nodes:
            self._update_path_node_flowrules(context, node, pc_corr,
                                             add_fc_ids, del_fc_ids)

    def _get_portchain_fcs(self, port_chain):
        return self._get_fcs_by_ids(port_chain['flow_classifiers'])

    def _get_fcs_by_ids(self, fc_ids):
        flow_classifiers = []
        if not fc_ids:
            return flow_classifiers

        # Get the portchain flow classifiers
        fc_plugin = (
            directory.get_plugin(flowclassifier.FLOW_CLASSIFIER_EXT)
        )
        if not fc_plugin:
            LOG.warning("Not found the flow classifier service plugin")
            return flow_classifiers

        for fc_id in fc_ids:
            fc = fc_plugin.get_flow_classifier(self.admin_context, fc_id)
            flow_classifiers.append(fc)

        return flow_classifiers

    @log_helpers.log_method_call
    def create_port_chain_precommit(self, context):
        """OVS Driver precommit before transaction committed.

        Make sure the logical_destination_port has been set when
        create symmetric port_chain
        """
        port_chain = context.current
        symmetric = port_chain['chain_parameters'].get('symmetric')
        if symmetric:
            for fc in self._get_fcs_by_ids(port_chain['flow_classifiers']):
                if fc['logical_destination_port'] is None:
                    raise exc.SfcBadRequest(message=(
                        'FlowClassifier %s does not set'
                        'logical_destination_port. logical_destination_port '
                        'needed when symmetric has been set. Please recreate '
                        'FlowClassifier with logical_destination_port and '
                        'destination_ip_prefix.' % fc['id']
                    ))

    @log_helpers.log_method_call
    def create_port_chain(self, context):
        port_chain = context.current
        symmetric = port_chain['chain_parameters'].get('symmetric')
        if symmetric:
            fwd_path_nodes = self._create_portchain_path(context, port_chain,
                                                         True)
            rev_path_nodes = self._create_portchain_path(context, port_chain,
                                                         False)
            self._update_path_nodes(
                context,
                fwd_path_nodes,
                port_chain['chain_parameters']['correlation'],
                port_chain['flow_classifiers'],
                None)
            self._update_path_nodes(
                context,
                rev_path_nodes,
                port_chain['chain_parameters']['correlation'],
                port_chain['flow_classifiers'],
                None)
        elif symmetric is False:
            path_nodes = self._create_portchain_path(context, port_chain, True)
            self._update_path_nodes(
                context,
                path_nodes,
                port_chain['chain_parameters']['correlation'],
                port_chain['flow_classifiers'],
                None)

    @log_helpers.log_method_call
    def delete_port_chain(self, context):
        port_chain = context.current
        LOG.debug("to delete portchain path")
        self._delete_portchain_path(context, port_chain)

    def _get_diff_set(self, orig, cur):
        orig_set = set(item for item in orig)
        cur_set = set(item for item in cur)

        to_del = orig_set.difference(cur_set)
        to_add = cur_set.difference(orig_set)

        return to_del, to_add

    @log_helpers.log_method_call
    def update_port_chain(self, context):
        port_chain = context.current
        orig = context.original
        self._delete_portchain_path(context, orig)
        # recreate port_chain after delete the orig
        symmetric = port_chain['chain_parameters'].get('symmetric')
        if symmetric:
            fwd_path_nodes = self._create_portchain_path(context, port_chain,
                                                         True)
            rev_path_nodes = self._create_portchain_path(context, port_chain,
                                                         False)
            self._update_path_nodes(
                context,
                fwd_path_nodes,
                port_chain['chain_parameters']['correlation'],
                port_chain['flow_classifiers'],
                None)
            self._update_path_nodes(
                context,
                rev_path_nodes,
                port_chain['chain_parameters']['correlation'],
                port_chain['flow_classifiers'],
                None)

        elif symmetric is False:
            path_nodes = self._create_portchain_path(context, port_chain, True)
            self._update_path_nodes(
                context,
                path_nodes,
                port_chain['chain_parameters']['correlation'],
                port_chain['flow_classifiers'],
                None)

    @log_helpers.log_method_call
    def create_port_pair_group(self, context):
        pass

    @log_helpers.log_method_call
    def delete_port_pair_group(self, context):
        pass

    @log_helpers.log_method_call
    def update_port_pair_group(self, context):
        current = context.current
        original = context.original

        if set(current['port_pairs']) == set(original['port_pairs']):
            return

        # Update the path_nodes and flows for each port chain that
        # contains this port_pair_group
        # Note: _get_port_pair_group is temporarily used here.
        ppg_obj = context._plugin._get_port_pair_group(context._plugin_context,
                                                       current['id'])
        port_chains = [assoc.portchain_id for assoc in
                       ppg_obj.chain_group_associations]

        for chain_id in port_chains:
            pc = context._plugin.get_port_chain(
                context._plugin_context, chain_id)
            pc_corr = pc['chain_parameters']['correlation']
            group_intid = current['group_id']
            # Get the previous node
            prev_nodes = self.get_path_nodes_by_filter(
                filters={'portchain_id': chain_id,
                         'next_group_id': group_intid})
            if not prev_nodes:
                continue

            for prev_node in prev_nodes:
                before_update_prev_node = prev_node.copy()
                # Update the previous node
                curr_group_intid, curr_group_members = \
                    self._get_portgroup_members(
                        context, current['id'], prev_node['fwd_path'])
                prev_node['next_hop'] = (
                    jsonutils.dumps(curr_group_members)
                    if curr_group_members else None
                )
                # update next hop to database
                self.update_path_node(prev_node['id'], prev_node)
                self._delete_path_node_flowrule(context,
                                                before_update_prev_node,
                                                pc_corr,
                                                pc['flow_classifiers'])
                self._update_path_node_flowrules(context,
                                                 prev_node,
                                                 pc_corr,
                                                 pc['flow_classifiers'],
                                                 None)

            # Update the current node
            # to find the current node by using the node's next_group_id
            # if this node is the last, next_group_id would be None
            curr_pos = pc['port_pair_groups'].index(current['id'])
            curr_nodes = self.get_path_nodes_by_filter(
                filters={'portchain_id': chain_id,
                         'nsi': 0xfe - curr_pos})

            if not curr_nodes:
                continue
            curr_node = None
            for temp_node in curr_nodes:
                if temp_node['fwd_path']:
                    curr_node = temp_node

            rev_curr_pos = len(pc['port_pair_groups']) - 1 - curr_pos
            rev_curr_nodes = self.get_path_nodes_by_filter(
                filters={'portchain_id': chain_id,
                         'nsi': 0xfe - rev_curr_pos})
            rev_curr_node = None
            if rev_curr_nodes is not None:
                for temp_node in rev_curr_nodes:
                    if temp_node['fwd_path'] is False:
                        rev_curr_node = temp_node

            # Add the port-pair-details into the current node
            for pp_id in (
                set(current['port_pairs']) - set(original['port_pairs'])
            ):
                ppd = self._get_port_pair_detail_by_port_pair(context,
                                                              pp_id)
                if not ppd:
                    LOG.debug('No port_pair_detail for the port_pair: %s',
                              pp_id)
                    LOG.debug("Failed to update port-pair-group")
                    return

                assco_args = {'portpair_id': ppd['id'],
                              'pathnode_id': curr_node['id'],
                              'weight': 1, }
                self.create_pathport_assoc(assco_args)
                updated_curr_node = self._update_tap_enabled_node(curr_node)
                self._update_path_node_port_flowrules(
                    context, updated_curr_node,
                    ppd, pc_corr, pc['flow_classifiers'])
                self._build_tap_ingress_flow(pc_corr=pc_corr,
                                             node=updated_curr_node,
                                             fwd_path=True)

                if not rev_curr_node:
                    continue
                assco_args = {'portpair_id': ppd['id'],
                              'pathnode_id': rev_curr_node['id'],
                              'weight': 1, }
                self.create_pathport_assoc(assco_args)
                updated_rev_curr_node = self._update_tap_enabled_node(
                    rev_curr_node)
                self._update_path_node_port_flowrules(context,
                                                      updated_rev_curr_node,
                                                      ppd,
                                                      pc_corr,
                                                      pc['flow_classifiers'])
                self._build_tap_ingress_flow(pc_corr=pc_corr,
                                             node=updated_rev_curr_node,
                                             fwd_path=False)

            # Delete the port-pair-details from the current node
            for pp_id in (
                set(original['port_pairs']) - set(current['port_pairs'])
            ):
                ppd = self._get_port_pair_detail_by_port_pair(context,
                                                              pp_id)
                if not ppd:
                    LOG.debug('No port_pair_detail for the port_pair: %s',
                              pp_id)
                    LOG.debug("Failed to update port-pair-group")
                    return
                updated_curr_node = self._update_tap_enabled_node(curr_node)
                self._delete_path_node_port_flowrule(
                    context,
                    updated_curr_node,
                    ppd,
                    pc_corr,
                    pc['flow_classifiers'])
                self.delete_pathport_assoc(updated_curr_node['id'], ppd['id'])

                if not rev_curr_node:
                    continue

                updated_rev_curr_node = self._update_tap_enabled_node(
                    rev_curr_node)
                self._delete_path_node_port_flowrule(
                    context,
                    updated_rev_curr_node,
                    ppd,
                    pc_corr,
                    pc['flow_classifiers'])
                self.delete_pathport_assoc(updated_rev_curr_node['id'],
                                           ppd['id'])

    @log_helpers.log_method_call
    def _get_port_detail_info(self, port_id):
        """Get port detail.

        @param: port_id: uuid
        @return: (host_id, local_ip, network_type, segment_id,
        service_insert_type): tuple
        """

        core_plugin = directory.get_plugin()
        port_detail = core_plugin.get_port(self.admin_context, port_id)
        host_id, local_ip, network_type, segment_id, mac_address = (
            (None, ) * 5)

        if port_detail:
            host_id = port_detail['binding:host_id']
            network_id = port_detail['network_id']
            mac_address = port_detail['mac_address']
            network_info = core_plugin.get_network(
                self.admin_context, network_id)
            network_type = network_info['provider:network_type']
            segment_id = network_info['provider:segmentation_id']

        if network_type != const.TYPE_VXLAN:
            LOG.warning("Currently only support vxlan network")
            return ((None, ) * 5)
        elif not host_id:
            LOG.warning("This port has not been binding")
            return ((None, ) * 5)
        else:
            driver = core_plugin.type_manager.drivers.get(network_type)
            host_endpoint = driver.obj.get_endpoint_by_host(host_id)
            if host_endpoint:
                local_ip = host_endpoint['ip_address']
            else:
                local_ip = None

        return host_id, local_ip, network_type, segment_id, mac_address

    @log_helpers.log_method_call
    def _create_port_pair_detail(self, port_pair):
        # since first node may not assign the ingress port, and last node
        # is not saved in the portpair_detail. we store the major egress port
        # info as the key to get the SF information.
        # mac_address stands for egress mac_address, and in_mac_address stands
        # for ingress_port mac_address
        in_port, e_port, host_id, local_endpoint, network_type, segment_id, \
            mac_address, in_mac_address = (
                (None, ) * 8)

        if port_pair.get('ingress', None):
            in_port = port_pair['ingress']
            in_host_id, in_local_endpoint, in_network_type, in_segment_id, \
                in_mac_address = (
                    self._get_port_detail_info(in_port))
        if port_pair.get('egress', None):
            e_port = port_pair['egress']
            host_id, local_endpoint, network_type, segment_id, mac_address = (
                self._get_port_detail_info(e_port))

        sfparams = port_pair.get('service_function_parameters')

        pp_corr = None
        if sfparams:
            pp_corr = sfparams.get('correlation')

        portpair_detail = {
            'ingress': port_pair.get('ingress', None),
            'egress': port_pair.get('egress', None),
            'correlation': pp_corr,
            'project_id': port_pair['project_id'],
            'host_id': host_id,
            'segment_id': segment_id,
            'network_type': network_type,
            'local_endpoint': local_endpoint,
            'mac_address': mac_address,
            'in_mac_address': in_mac_address
        }
        r = self.create_port_pair_detail(portpair_detail)
        LOG.debug('create port-pair detail: %s', r)
        return r

    @log_helpers.log_method_call
    def create_port_pair(self, context):
        port_pair = context.current
        self._create_port_pair_detail(port_pair)

    @log_helpers.log_method_call
    def delete_port_pair(self, context):
        port_pair = context.current

        pd_filter = dict(ingress=port_pair.get('ingress', None),
                         egress=port_pair.get('egress', None),
                         project_id=port_pair['project_id']
                         )
        pds = self.get_port_details_by_filter(pd_filter)
        if pds:
            for pd in pds:
                self.delete_port_pair_detail(pd['id'])

    @log_helpers.log_method_call
    def update_port_pair(self, context):
        pass

    def get_flowrules_by_host_portid(self, context, host, port_id):
        port_chain_flowrules = []
        sfc_plugin = directory.get_plugin(sfc.SFC_EXT)
        if not sfc_plugin:
            return port_chain_flowrules
        try:
            port_detail_list = []
            # one port only may be in egress/ingress port once time.
            ingress_port = self.get_port_detail_by_filter(
                dict(ingress=port_id))
            egress_port = self.get_port_detail_by_filter(
                dict(egress=port_id))
            if not ingress_port and not egress_port:
                return None
            # SF migrate to other host
            if ingress_port:
                port_detail_list.append(ingress_port)
                if ingress_port['host_id'] != host:
                    ingress_port.update(dict(host_id=host))

            if egress_port:
                port_detail_list.append(egress_port)
                if egress_port['host_id'] != host:
                    egress_port.update(dict(host_id=host))

            # Get ingress flow for Tap SF
            if self._is_tap_ports(port_detail_list):
                return self._get_tap_port_ingress_flow(context,
                                                       sfc_plugin,
                                                       port_chain_flowrules,
                                                       port_detail_list)

            # this is a SF if there are both egress and ingress.
            for i, port in enumerate(port_detail_list):
                nodes_assocs = port['path_nodes']
                for assoc in nodes_assocs:
                    # update current path flow rule
                    node = self.get_path_node(assoc['pathnode_id'])
                    port_chain = sfc_plugin.get_port_chain(
                        context,
                        node['portchain_id'])

                    # if this path_node is a linking node of a graph,
                    # then we obtain the respective flow_rule
                    flow_rule = self._get_flow_rule_for_service_graph_node(
                        context, node, port, port_chain)

                    # otherwise it's a normal port chain flow_rule
                    if not flow_rule:
                        flow_rule = self._build_portchain_flowrule_body(
                            node,
                            port,
                            port_chain['chain_parameters']['correlation'],
                            add_fc_ids=port_chain['flow_classifiers']
                        )
                    port_chain_flowrules.append(flow_rule)
            self._get_egress_flowrule_if_next_hop_tap(sfc_plugin,
                                                      context,
                                                      port_chain_flowrules,
                                                      port_detail_list)

            return port_chain_flowrules

        except Exception as e:
            LOG.exception(e)
            LOG.error("get_flowrules_by_host_portid failed")

    def _get_flow_rule_for_service_graph_node(
            self, context, node, port, port_chain):
        correlation = port_chain['chain_parameters']['correlation']
        sfc_plugin = directory.get_plugin(sfc.SFC_EXT)
        if node['node_type'] == ovs_const.SRC_NODE:
            # check if port_chain is a dst_chain in a graph
            branches = sfc_plugin._get_branches(
                context, filters={'dst_chain': [port_chain['id']]})
            if branches:
                matches = set()
                for branch in branches:
                    src_chain = branch['src_chain']
                    # bnodes - "branching" nodes
                    bnodes = self.get_path_nodes_by_filter(
                        filters={
                            'portchain_id': src_chain,
                            'node_type': ovs_const.SF_NODE,
                            'next_hops': None,
                            'next_group_ids': None})
                    # after filtering on the database, we still have to
                    # filter on the singular next_group_id and next_hop by code
                    bnodes = [x for x in bnodes if x[
                        'next_group_id'] is None and x['next_hop'] is None]
                    if not bnodes:
                        LOG.error(
                            '1 branching path node was expected, '
                            'but none were found.')
                    elif len(bnodes) != 1:
                        LOG.error(
                            '1 branching path node was expected, '
                            'but %d were found.', len(bnodes))
                    nsp = bnodes[0]['nsp']
                    nsi = bnodes[0]['nsi']
                    matches.add((nsp, nsi,))
                flow_rule = self._build_bare_flow_rule(
                    node, port, correlation, correlation)
                add_fcs, del_fcs = self._build_fcs_from_fc_ids(
                    flow_rule, port_chain['flow_classifiers'])
                return self._extend_sg_dst_chain_flow_rule(
                    flow_rule, True, matches, add_fcs, [])
        elif (node['node_type'] == ovs_const.SF_NODE and
              node['next_hop'] is None and
              node['next_group_id'] is None):
            # check if port_chain is a src_chain in a graph
            branches = sfc_plugin._get_branches(
                context, filters={'src_chain': [port_chain['id']]})
            if branches:
                flow_rule = self._build_bare_flow_rule(
                    node, port, correlation, correlation)
                add_fcs, del_fcs = self._build_fcs_from_fc_ids(
                    flow_rule, port_chain['flow_classifiers'])
                return self._extend_sg_src_chain_flow_rule(
                    flow_rule, True, add_fcs, [])
        return None

    def _build_bare_flow_rule(self, node, port, pc_corr, pp_corr):
        node_info = node.copy()
        node_info.pop('project_id')
        node_info.pop('portpair_details')
        port_info = port.copy()
        port_info.pop('project_id')
        port_info.pop('id')
        port_info.pop('path_nodes')
        flow_rule = dict(node_info, **port_info)
        flow_rule['pc_corr'] = pc_corr
        flow_rule['pp_corr'] = pc_corr
        # correlation becomes simply pp_corr
        flow_rule.pop('correlation')
        return flow_rule

    def _build_fcs_from_fc_ids(self, flow_rule, fc_ids):
        fcs = self._filter_flow_classifiers(flow_rule, fc_ids)
        add_fcs = []
        del_fcs = []
        for fc in fcs:
            add_fcs.append(fc)
            del_fcs.append(fc)
        return add_fcs, del_fcs

    def _extend_sg_src_chain_flow_rule(
            self, flow_rule, branch_point, add_fcs, del_fcs):
        # trigger chain-linking
        flow_rule['branch_point'] = branch_point
        flow_rule['add_fcs'] = add_fcs
        flow_rule['del_fcs'] = del_fcs
        return flow_rule

    def _extend_sg_dst_chain_flow_rule(
            self, flow_rule, on_add, matches, add_fcs=None, del_fcs=None):
        if add_fcs is not None:
            flow_rule['add_fcs'] = add_fcs
        if del_fcs is not None:
            flow_rule['del_fcs'] = del_fcs
        flow_rule['branch_info'] = {}
        flow_rule['branch_info']['on_add'] = on_add
        flow_rule['branch_info']['matches'] = matches
        return flow_rule

    def update_flowrule_status(self, context, id, status):
        """FIXME

        drivers/ovs/db.py will be removed in the future with 4 ovs tables

        This function raise:
            RuntimeError: reentrant call
            DBError: reentrant call
            DBConnectionError: (pymysql.err.OperationalError)
                (2014, 'Command Out of Sync'))
        """
        pass
        # try:
        #     flowrule_status = dict(status=status)
        #     self.update_path_node(id, flowrule_status)
        # except Exception as e:
        #     LOG.exception(e)
        #     LOG.error("update_flowrule_status failed")

    def _update_portchain_group_reference_count(self, flow_rule, host):
        group_refcnt = 0
        flow_rule['host'] = host

        if flow_rule['next_group_id'] is not None:
            all_nodes = self.get_path_nodes_by_filter(
                filters={'next_group_id': flow_rule['next_group_id'],
                         'nsi': 0xff})
            if all_nodes is not None:
                for node in all_nodes:
                    if not node['portpair_details']:
                        if flow_rule['fwd_path'] == node['fwd_path']:
                            group_refcnt += 1

            port_details = self.get_port_details_by_filter(
                dict(host_id=flow_rule['host']))
            if port_details is not None:
                for pd in port_details:
                    for path in pd['path_nodes']:
                        path_node = self.get_path_node(path['pathnode_id'])
                        if (
                            path_node['next_group_id'] ==
                            flow_rule['next_group_id'] and
                            path_node['fwd_path']
                        ):
                            group_refcnt += 1
        flow_rule['group_refcnt'] = group_refcnt

        return group_refcnt

    def _service_graph_linking_logic(self, context, create):
        service_graph = context.current
        # keep track of matches per dst chain (key)
        dc_branch_info = {}
        for src_chain_id in service_graph['port_chains']:
            sc_nodes = self.get_path_nodes_by_filter(
                dict(portchain_id=src_chain_id))
            sc_nodes = sorted(sc_nodes, key=lambda n: n['nsi'])
            # dst_node is ignored when linking a chain to another chain
            if sc_nodes[0]['node_type'] == 'dst_node':
                del sc_nodes[0]
            # one of branch_info (it's for the current src chain),
            # sc_nodes[0] is simply the last sf_node the src chain
            branch_match = (sc_nodes[0]['nsp'], sc_nodes[0]['nsi'],)
            branches = 0
            # for clarification: sc = source chain; dc = destination chain
            for dst_chain_id in service_graph['port_chains'][src_chain_id]:
                dc_nodes = self.get_path_nodes_by_filter(
                    dict(portchain_id=dst_chain_id))
                if dst_chain_id not in dc_branch_info:
                    dc_branch_info[dst_chain_id] = {}
                    dc_branch_info[dst_chain_id]['matches'] = []
                    dc_branch_info[dst_chain_id]['flow_rules'] = {}
                # trigger chain-linking using match of current src chain
                dc_branch_info[dst_chain_id]['matches'].append(branch_match)
                for node in dc_nodes:
                    if node['node_type'] == ovs_const.SRC_NODE:
                        dst_chain = context._plugin.get_port_chain(
                            context._plugin_context, dst_chain_id)
                        pc_corr = dst_chain[
                            'chain_parameters']['correlation']
                        new_fcs = dst_chain['flow_classifiers']
                        for each in node['portpair_details']:
                            port = self.get_port_detail_by_filter(
                                dict(id=each))
                            node_info = node.copy()
                            node_info.pop('project_id')
                            node_info.pop('portpair_details')
                            port_info = port.copy()
                            port_info.pop('project_id')
                            port_info.pop('id')
                            port_info.pop('path_nodes')

                            flow_rule = dict(node_info, **port_info)
                            flow_rule['pc_corr'] = pc_corr
                            # there's no correlation for src nodes
                            flow_rule['pp_corr'] = None
                            # correlation becomes simply pp_corr
                            flow_rule.pop('correlation')

                            old_fcs = self._filter_flow_classifiers(flow_rule,
                                                                    new_fcs)
                            for old_fc in old_fcs:
                                new_fc = old_fc.copy()
                                if create:
                                    # dependent PCs should ignore LSPs
                                    new_fc['logical_source_port'] = None
                                else:
                                    old_fc['logical_source_port'] = None
                                if 'del_fcs' not in flow_rule:
                                    flow_rule['del_fcs'] = []
                                if 'add_fcs' not in flow_rule:
                                    flow_rule['add_fcs'] = []
                                flow_rule['del_fcs'].append(old_fc)
                                flow_rule['add_fcs'].append(new_fc)

                            # update next hop info
                            self._update_path_node_next_hops(flow_rule)
                            # usually we would ask_agent_to_update_flow_rules
                            # here but "joining" branches require postponing
                            dc_branch_info[
                                dst_chain_id]['flow_rules'][each] = flow_rule
                            branches = branches + 1

            if branches > 0:  # this is supposed to always be true
                node = sc_nodes[0]
                if node['node_type'] == ovs_const.SF_NODE:
                    src_chain = context._plugin.get_port_chain(
                        context._plugin_context, src_chain_id)
                    fc_ids = src_chain['flow_classifiers']
                    pc_corr = src_chain['chain_parameters']['correlation']
                    for each in node['portpair_details']:
                        port = self.get_port_detail_by_filter(dict(id=each))
                        node_info = node.copy()
                        port_info = port.copy()
                        flow_rule = self._build_bare_flow_rule(
                            node_info, port_info, pc_corr, pc_corr)
                        add_fcs, del_fcs = self._build_fcs_from_fc_ids(
                            flow_rule, fc_ids)
                        flow_rule = self._extend_sg_src_chain_flow_rule(
                            flow_rule, create, add_fcs, del_fcs)
                        # update next hop info
                        self._update_path_node_next_hops(flow_rule)
                        # call agent
                        self.ovs_driver_rpc.ask_agent_to_update_flow_rules(
                            self.admin_context,
                            flow_rule)
                        self._update_agent_fdb_entries(context, flow_rule)
            else:
                LOG.warning('Expected branch point did not get any branches.')
        # at this point we know about all chain links/matches
        # (including "joining" branches), so let's call the agent
        for dst_chain_id in dc_branch_info:
            branch_matches = dc_branch_info[dst_chain_id]['matches']
            flow_rules = dc_branch_info[dst_chain_id]['flow_rules']
            for _, flow_rule in flow_rules.items():
                flow_rule = self._extend_sg_dst_chain_flow_rule(
                    flow_rule, create, branch_matches)
                # call agent
                self.ovs_driver_rpc.ask_agent_to_update_flow_rules(
                    self.admin_context,
                    flow_rule)
                self._update_agent_fdb_entries(context, flow_rule)

    @log_helpers.log_method_call
    def create_service_graph_precommit(self, context):
        pass

    @log_helpers.log_method_call
    def create_service_graph_postcommit(self, context):
        self._service_graph_linking_logic(context, True)

    @log_helpers.log_method_call
    def update_service_graph_precommit(self, context):
        pass

    @log_helpers.log_method_call
    def update_service_graph_postcommit(self, context):
        pass

    @log_helpers.log_method_call
    def delete_service_graph_precommit(self, context):
        pass

    @log_helpers.log_method_call
    def delete_service_graph_postcommit(self, context):
        self._service_graph_linking_logic(context, False)

    def _get_tap_port_ingress_flow(self, context, sfc_plugin,
                                   port_chain_flowrules, port_detail_list):
        for i, ports in enumerate(port_detail_list):
            nodes_assocs = ports['path_nodes']
            for assoc in nodes_assocs:
                node = self.get_path_node(assoc['pathnode_id'])
                updated_node = self._update_tap_enabled_node(node)
                port_chain = sfc_plugin.get_port_chain(
                    context,
                    node['portchain_id'])
                flowrules = self._build_tap_ingress_flow(
                    port_chain['chain_parameters']['correlation'],
                    updated_node,
                    node['fwd_path'],
                    configure=False)
                port_chain_flowrules.extend(flowrules)
        return port_chain_flowrules

    def _get_egress_flowrule_if_next_hop_tap(self, sfc_plugin, context,
                                             port_chain_flowrules,
                                             port_detail_list):
        # Get the flow for Tap egress port.
        for i, ports in enumerate(port_detail_list):
            nodes_assocs = ports['path_nodes']
            for assoc in nodes_assocs:
                node = self.get_path_node(assoc['pathnode_id'])
                if not node['next_hop']:
                    return
                next_hops = jsonutils.loads(node['next_hop'])
                for next_hop in next_hops:
                    if not next_hop['tap_enabled']:
                        return

                    tap_pp = self.get_port_detail_by_filter(
                        dict(id=next_hop['portpair_id']))
                    for t_node in tap_pp['path_nodes']:
                        tap_node = self.get_path_node(t_node['pathnode_id'])
                        tap_node = self._update_tap_enabled_node(tap_node)
                        if tap_node['fwd_path'] == node['fwd_path']:
                            port_chain = sfc_plugin.get_port_chain(
                                context,
                                tap_node['portchain_id'])
                            flowrule = self._build_portchain_flowrule_body(
                                tap_node,
                                ports,
                                port_chain['chain_parameters']['correlation'],
                                add_fc_ids=port_chain['flow_classifiers']
                            )
                            port_chain_flowrules.append(flowrule)

    def _validate_chain(self, context, port_chain, port_pair_groups, fwd_path):
        """Validate the cross-subnet restriction

        :param context:
        :param port_chain:
        :param port_pair_groups:
        :return:
        """
        sf_path_length = len(port_pair_groups)
        ppgs = [context._plugin.get_port_pair_group(context._plugin_context,
                ppg) for ppg in port_pair_groups]

        # Detect cross-subnet transit
        # Compare subnets for logical source ports
        # and first PPG ingress ports
        for fc in self._get_fcs_by_ids(port_chain['flow_classifiers']):
            subnet1 = self._get_subnet_by_port(fc['logical_source_port'])
            cidr1 = subnet1['cidr']
            ppg = ppgs[0]
            for pp_id1 in ppg['port_pairs']:
                pp1 = context._plugin.get_port_pair(context._plugin_context,
                                                    pp_id1)
                filter1 = {}
                if pp1.get('ingress', None):
                    filter1 = dict(dict(ingress=pp1['ingress']), **filter1)
                    pd1 = self.get_port_detail_by_filter(filter1)
                    subnet2 = self._get_subnet_by_port(pd1['ingress'])
                    cidr2 = subnet2['cidr']
                    if cidr1 != cidr2:
                        LOG.error('Cross-subnet chain not supported')
                        raise exc.SfcDriverError(
                            method='create_portchain_path')

        # Compare subnets for PPG egress ports
        # and next PPG ingress ports
        for i in range(sf_path_length - 1):
            ppg = ppgs[i]
            next_ppg = ppgs[i + 1]
            for pp_id1 in ppg['port_pairs']:
                pp1 = context._plugin.get_port_pair(context._plugin_context,
                                                    pp_id1)
                filter1 = {}
                cidr3 = None
                if pp1.get('egress', None):
                    filter1 = dict(dict(egress=pp1['egress']), **filter1)
                    pd1 = self.get_port_detail_by_filter(filter1)
                    subnet1 = self._get_subnet_by_port(pd1['egress'])
                    cidr3 = subnet1['cidr']

                for pp_id2 in next_ppg['port_pairs']:
                    pp2 = context._plugin.get_port_pair(
                        context._plugin_context, pp_id2)
                    filter2 = {}
                    if pp2.get('ingress', None):
                        filter2 = dict(dict(ingress=pp2['ingress']), **filter2)
                        pd2 = self.get_port_detail_by_filter(filter2)
                        subnet2 = self._get_subnet_by_port(pd2['ingress'])
                        cidr4 = subnet2['cidr']
                        if cidr3 != cidr4:
                            LOG.error('Cross-subnet chain not supported')
                            raise exc.SfcDriverError(
                                method='create_portchain_path')

        ppgs = ppgs if fwd_path else list(reversed(ppgs))
        return ppgs

    def _update_tap_enabled_node(self, node):
        if not node['tap_enabled'] or not node['previous_node_id']:
            return node

        prev_node = self.get_path_node(node['previous_node_id'])
        node_copy = node.copy()
        node_copy['prev_node'] = prev_node
        node_copy['tap_node'] = node
        node_copy['tap_pp_corr'] = self.get_port_detail(
            node['portpair_details'][0])['correlation']
        node_copy.update(id=prev_node['id'],
                         nsi=prev_node['nsi'],
                         nsp=prev_node['nsp'],
                         portpair_details=prev_node['portpair_details'],
                         node_type=prev_node['node_type'],
                         skip_ingress_flow_config=True)  # Flags driver to
        # skip ingress flow configuration.
        return node_copy

    def _build_tap_ingress_flow(self, pc_corr, node, fwd_path, delete=False,
                                configure=True):

        """Build add/delete ingress flows of Tap node.

        :param pc_corr:
        :param node:
        :param fwd_path:
        :param delete:
        :param configure:
        :return:
        """
        if not node['tap_enabled']:
            return
        flowrule_list = list()
        # 'fwd_path' is always set to true. Call from this method to OVS agent
        # is made to either create/delete ingress flows explicitly. In ref.
        # sfc_driver.py - L134 swap ingress & egress flows if fwd_path set
        # to False, which is not required for Tap SF's ingress flows.
        flowrule = {'nsp': node['tap_node']['nsp'],
                    'nsi': node['tap_node']['nsi'],
                    'node_nsp': node['nsp'],
                    'node_nsi': node['nsi'],
                    'pc_corr': pc_corr,
                    'pp_corr': node['tap_pp_corr'],
                    'pp_corr_tap_nh': None,
                    'id': node['id'],
                    'fwd_path': True,
                    'node_type': node['node_type'],
                    'tap_nh_node_type': None,
                    'portchain_id': node['portchain_id'],
                    }

        if node['tap_node']['next_hop']:
            next_hop = jsonutils.loads(node['tap_node']['next_hop'])
            nh_pp = self.get_port_detail(next_hop[0]['portpair_id'])
            flowrule.update(pp_corr_tap_nh=nh_pp['correlation'],
                            tap_nh_node_type=ovs_const.SF_NODE)

        for each in node['prev_node']['portpair_details']:
            port = self.get_port_detail_by_filter(dict(id=each))
            if not port:
                continue
            if fwd_path:
                flowrule.update(mac_address=port['mac_address'])
            else:
                flowrule.update(
                    mac_address=(port['in_mac_address'] or
                                 port['mac_address']))

            for pp_id in node['tap_node']['portpair_details']:
                _port = self.get_port_detail_by_filter(dict(id=pp_id))
                if _port:
                    # when Tap is launched with single port
                    ingress_port = _port['ingress'] if fwd_path else (
                        _port['egress'] or _port['ingress'])
                    flowrule.update(egress=None,
                                    ingress=ingress_port,
                                    host=_port['host_id'],
                                    pp_corr=_port['correlation'],
                                    tap_enabled=True)
                    flowrule_list.append(flowrule.copy())

        if configure:
            self._update_tap_ingress_flow_rules(flowrule_list, delete)
        else:
            return flowrule_list

    def _update_tap_ingress_flow_rules(self, flowrule_list, delete):
        for flowrule in flowrule_list:
            if not delete:
                self.ovs_driver_rpc.ask_agent_to_update_flow_rules(
                    self.admin_context,
                    flowrule)
            else:
                self.ovs_driver_rpc.ask_agent_to_delete_flow_rules(
                    self.admin_context,
                    flowrule
                )

    def _is_tap_ports(self, port_detail_list):
        for port in port_detail_list:
            nodes_assocs = port['path_nodes']
            for assoc in nodes_assocs:
                node = self.get_path_node(assoc['pathnode_id'])
                if node['tap_enabled']:
                    return True
        return False

    def _update_sc_path_parameter(self, nodes):
        # This method updates 'nsi' of nodes in service chain. 'nsi' is
        # assigned to nodes in decreasing order. For eg., if the nodes are
        # like: SRC_NODE -> TAP_NODE -> DEFAULT_NODE -> DST_NODE - 'nsi' for
        # corresponding nodes will be - 255 -> 254 -> 253 -> 252 , where 254
        # is of TAP_NODE. In this method, TAP_NODES 'nsi' are re-assigned
        # from the DST_NODE onward, for eg. the new TAP_NODE 'nsi' would be
        # 252 instead of 254, the resultant chain's 'nsi' - SRC_NODE (255) ->
        # TAP_NODE (252) -> DEFAULT_NODE (254) -> DST_NODE (253) . This was
        # done to simplify the OVS driver configuration, with the various
        # combination of correlation parameters of port-pairs.
        tap_sf_in_chain = False
        def_nodes, tap_nodes = list(), list()
        node_list = [None] * len(nodes)
        for node in nodes:
            node_list[0xff - node['nsi']] = node
        for node in node_list:
            if node['tap_enabled']:
                tap_sf_in_chain = True
                tap_nodes.append(node)
            else:
                def_nodes.append(node)
        if not tap_sf_in_chain:
            return nodes
        else:
            node_list = [def_nodes[0]]
            for i, node in enumerate(def_nodes):
                if i > 0:
                    node_list.append(self.update_path_node(
                        node['id'], dict(nsi=def_nodes[0]['nsi'] - i)))
            tap_nsi = def_nodes[0]['nsi'] - len(def_nodes)
            for node in tap_nodes:
                node_list.append(self.update_path_node(node['id'],
                                                       dict(nsi=tap_nsi)))
                tap_nsi += 1
            return node_list

    def _update_next_hop_details(self, flow_rule, port_detail, detail):
        # Update detail with next-hop node_type and the correlation of
        # port-pair of next-hop
        for node in port_detail['path_nodes']:
            _node = self.get_path_node(node['pathnode_id'])
            if flow_rule['fwd_path'] == _node['fwd_path']:
                detail['nsi'], detail['nsp'] = _node['nsi'], _node['nsp']
                if _node['next_hop']:
                    next_hops = jsonutils.loads(_node['next_hop'])
                    for next_hop in next_hops:
                        pp = self.get_port_detail(next_hop['portpair_id'])
                        detail['pp_corr_tap_nh'] = pp['correlation']
                        for pp_node in pp['path_nodes']:
                            pp_node_detail = self.get_path_node(
                                pp_node['pathnode_id'])
                            detail['tap_nh_node_type'] = pp_node_detail[
                                'node_type']
                            return



# Copyright 2015 Huawei.
# Copyright 2016 Red Hat, Inc.
# Copyright 2017 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import constants as n_consts
from oslo_config import cfg
from oslo_log import log as logging

from neutron.plugins.ml2.drivers.openvswitch.agent.common import constants \
    as ovs_consts
from neutron.plugins.ml2.drivers.openvswitch.agent import vlanmanager

from networking_sfc.services.sfc.agent.extensions import sfc
from networking_sfc.services.sfc.common import ovs_ext_lib
from networking_sfc.services.sfc.drivers.ovs import constants

LOG = logging.getLogger(__name__)

cfg.CONF.import_group('OVS', 'neutron.plugins.ml2.drivers.openvswitch.agent.'
                             'common.config')

# This table is used to process the traffic across differet subnet scenario.
# Flow 1: pri=1, ip,dl_dst=nexthop_mac,nw_src=nexthop_subnet. actions=
# push_mpls:0x8847,set_mpls_label,set_mpls_ttl,push_vlan,output:(patch port
# or resubmit to table(INGRESS_TABLE)
# Flow 2: pri=0, ip,dl_dst=nexthop_mac,, action=push_mpls:0x8847,
# set_mpls_label,set_mpls_ttl,push_vlan,output:(patch port or resubmit to
# table(INGRESS_TABLE)
ACROSS_SUBNET_TABLE = 5

# The table has multiple flows that steer traffic for the different chains
# to the ingress port of different service functions hosted on this Compute
# node.
INGRESS_TABLE = 10

# port chain default flow rule priority
PC_DEF_PRI = 20
PC_INGRESS_PRI = 30

# Reverse group number offset for dump_group
REVERSE_GROUP_NUMBER_OFFSET = 7000

TAP_CLASSIFIER_TABLE = 7
# This table floods TAP packets on tunnel ports
TAP_TUNNEL_OUTPUT_TABLE = 25

# actions
RESUBMIT_TAP_TABLE = ',resubmit(,%s)' % TAP_CLASSIFIER_TABLE
NORMAL_ACTION = ",NORMAL"


class SfcOVSAgentDriver(sfc.SfcAgentDriver):
    """This class will support MPLS frame

    Ethernet + MPLS
    IPv4 Packet:
    +-------------------------------+---------------+--------------------+
    |Outer Ethernet, ET=0x8847      | MPLS head,    | original IP Packet |
    +-------------------------------+---------------+--------------------+
    """

    def __init__(self):
        super(SfcOVSAgentDriver, self).__init__()
        self.agent_api = None
        self.br_int = None
        self.br_tun = None

        self.local_ip = None
        self.patch_tun_ofport = None
        self.vlan_manager = None

    def consume_api(self, agent_api):
        self.agent_api = agent_api

    def initialize(self):
        self.br_int = ovs_ext_lib.SfcOVSBridgeExt(
            self.agent_api.request_int_br())

        self.local_ip = cfg.CONF.OVS.local_ip
        self.patch_tun_ofport = self.br_int.get_port_ofport(
            cfg.CONF.OVS.int_peer_patch_port)
        self.vlan_manager = vlanmanager.LocalVlanManager()

        self._clear_sfc_flow_on_int_br()

        self.br_tun = ovs_ext_lib.SfcOVSBridgeExt(
            self.agent_api.request_tun_br())
        self.patch_int_ofport = self.br_tun.get_port_ofport(
            cfg.CONF.OVS.tun_peer_patch_port)
        self._clear_sfc_flow_on_tun_br()

    def update_flow_rules(self, flowrule, flowrule_status):
        if flowrule['fwd_path'] is False and flowrule['node_type'] == \
                'sf_node':
            flowrule['ingress'], flowrule['egress'] = flowrule['egress'], \
                flowrule['ingress']
        try:
            LOG.debug('update_flow_rule, flowrule = %s', flowrule)

            if flowrule.get('egress', None):
                self._setup_egress_flow_rules(flowrule)
            if flowrule.get('ingress', None) and not flowrule.get(
                    'skip_ingress_flow_config', None):
                self._setup_ingress_flow_rules(flowrule)
            flowrule_status_temp = {'id': flowrule['id'],
                                    'status': constants.STATUS_ACTIVE}
            flowrule_status.append(flowrule_status_temp)
        except Exception as e:
            flowrule_status_temp = {'id': flowrule['id'],
                                    'status': constants.STATUS_ERROR}
            flowrule_status.append(flowrule_status_temp)
            LOG.exception(e)
            LOG.error("update_flow_rules failed")

    def delete_flow_rule(self, flowrule, flowrule_status):
        if flowrule['fwd_path'] is False and flowrule['node_type'] == \
                'sf_node':
            flowrule['ingress'], flowrule['egress'] = flowrule['egress'], \
                flowrule['ingress']
        try:
            LOG.debug("delete_flow_rule, flowrule = %s", flowrule)
            pc_corr = flowrule['pc_corr']

            # delete tunnel table flow rule on br-int(egress match)
            if flowrule['egress'] is not None:
                self._setup_local_switch_flows_on_int_br(flowrule,
                                                         flowrule['del_fcs'],
                                                         None,
                                                         add_flow=False,
                                                         match_inport=True)
                # delete group table, need to check again
                group_id = flowrule.get('next_group_id', None)
                if group_id and flowrule.get('group_refcnt', None) <= 1:
                    if flowrule['fwd_path']:
                        self.br_int.delete_group(group_id=group_id)
                    else:
                        self.br_int.delete_group(group_id=group_id +
                                                 REVERSE_GROUP_NUMBER_OFFSET)
                    self._delete_across_subnet_table_flows(flowrule)

            if flowrule['ingress'] is not None:
                # delete table INGRESS_TABLE ingress match flow rule
                # on br-int(ingress match)
                vif_port = self.br_int.get_vif_port_by_id(flowrule['ingress'])
                if vif_port:
                    # third, install br-int flow rule on table INGRESS_TABLE
                    # for ingress traffic
                    if pc_corr == 'mpls':
                        self._delete_flows_mpls(flowrule, vif_port)
                    elif pc_corr == 'nsh':
                        self._delete_flows_nsh(flowrule, vif_port)
        except Exception as e:
            flowrule_status_temp = {'id': flowrule['id'],
                                    'status': constants.STATUS_ERROR}
            flowrule_status.append(flowrule_status_temp)
            LOG.exception(e)
            LOG.error("delete_flow_rule failed")

    def _clear_sfc_flow_on_int_br(self):
        self.br_int.delete_group(group_id='all')
        self.br_int.delete_flows(table=ACROSS_SUBNET_TABLE)
        self.br_int.delete_flows(table=INGRESS_TABLE)
        self.br_int.delete_flows(table=TAP_CLASSIFIER_TABLE)
        self.br_int.install_goto(dest_table_id=INGRESS_TABLE,
                                 priority=PC_DEF_PRI,
                                 eth_type=constants.ETH_TYPE_MPLS)
        self.br_int.install_goto(dest_table_id=INGRESS_TABLE,
                                 priority=PC_DEF_PRI,
                                 eth_type=constants.ETH_TYPE_NSH)
        self.br_int.install_drop(table_id=INGRESS_TABLE)

    def _parse_flow_classifier(self, flow_classifier):
        eth_type, nw_proto, source_port_masks, destination_port_masks = (
            (None, ) * 4)

        if (not flow_classifier['source_port_range_min'] and
                not flow_classifier['source_port_range_max']):
            # wildcard
            source_port_masks = ['0/0x0']
        elif not flow_classifier['source_port_range_min']:
            source_port_masks = ovs_ext_lib.get_port_mask(
                1,
                flow_classifier['source_port_range_max'])
        elif not flow_classifier['source_port_range_max']:
            source_port_masks = ovs_ext_lib.get_port_mask(
                flow_classifier['source_port_range_min'],
                65535)
        else:
            source_port_masks = ovs_ext_lib.get_port_mask(
                flow_classifier['source_port_range_min'],
                flow_classifier['source_port_range_max'])

        if (not flow_classifier['destination_port_range_min'] and
                not flow_classifier['destination_port_range_max']):
            # wildcard
            destination_port_masks = ['0/0x0']
        elif not flow_classifier['destination_port_range_min']:
            destination_port_masks = ovs_ext_lib.get_port_mask(
                1,
                flow_classifier['destination_port_range_max'])
        elif not flow_classifier['destination_port_range_max']:
            destination_port_masks = ovs_ext_lib.get_port_mask(
                flow_classifier['destination_port_range_min'],
                65535)
        else:
            destination_port_masks = ovs_ext_lib.get_port_mask(
                flow_classifier['destination_port_range_min'],
                flow_classifier['destination_port_range_max'])

        if flow_classifier['ethertype'] == "IPv4":
            eth_type = constants.ETH_TYPE_IP
            if n_consts.PROTO_NAME_TCP == flow_classifier['protocol']:
                nw_proto = n_consts.PROTO_NUM_TCP
            elif n_consts.PROTO_NAME_UDP == flow_classifier['protocol']:
                nw_proto = n_consts.PROTO_NUM_UDP
            elif n_consts.PROTO_NAME_ICMP == flow_classifier['protocol']:
                nw_proto = n_consts.PROTO_NUM_ICMP
            else:
                nw_proto = None
        elif flow_classifier['ethertype'] == "IPv6":
            LOG.error("Current portchain agent doesn't support IPv6")
        else:
            LOG.error("invalid protocol input")

        return (eth_type, nw_proto,
                source_port_masks,
                destination_port_masks)

    def _get_flow_infos_from_flow_classifier(self, flow_classifier, flowrule):
        flow_infos = []
        nw_src, nw_dst, tp_src, tp_dst = ((None, ) * 4)

        if flow_classifier['ethertype'] != "IPv4":
            LOG.error("Current portchain agent only supports IPv4")
            return flow_infos

        # parse and transfer flow info to match field info
        eth_type, nw_proto, source_port_masks, destination_port_masks = (
            self._parse_flow_classifier(flow_classifier))

        if flowrule['fwd_path']:
            if flow_classifier['source_ip_prefix']:
                nw_src = flow_classifier['source_ip_prefix']
            else:
                nw_src = '0.0.0.0/0.0.0.0'
            if flow_classifier['destination_ip_prefix']:
                nw_dst = flow_classifier['destination_ip_prefix']
            else:
                nw_dst = '0.0.0.0/0.0.0.0'
        else:
            if flow_classifier['source_ip_prefix']:
                nw_src = flow_classifier['destination_ip_prefix']
            else:
                nw_src = '0.0.0.0/0.0.0.0'
            if flow_classifier['destination_ip_prefix']:
                nw_dst = flow_classifier['source_ip_prefix']
            else:
                nw_dst = '0.0.0.0/0.0.0.0'

        if source_port_masks and destination_port_masks:
            for destination_port in destination_port_masks:
                for source_port in source_port_masks:
                    if flowrule['fwd_path']:
                        tp_src = '%s' % source_port
                        tp_dst = '%s' % destination_port
                    else:
                        tp_dst = '%s' % source_port
                        tp_src = '%s' % destination_port
                    flow_info = {'eth_type': eth_type,
                                 'nw_src': nw_src,
                                 'nw_dst': nw_dst,
                                 'tp_src': tp_src,
                                 'tp_dst': tp_dst}
                    if nw_proto:
                        flow_info['nw_proto'] = nw_proto
                    flow_infos.append(flow_info)

        return flow_infos

    def _get_flow_infos_from_flow_classifier_list(self, flow_classifier_list,
                                                  flowrule):
        flow_infos = []

        if not flow_classifier_list:
            return flow_infos
        for flow_classifier in flow_classifier_list:
            flow_infos.extend(
                self._get_flow_infos_from_flow_classifier(flow_classifier,
                                                          flowrule))

        return flow_infos

    def _match_by_header(self, match_info, nsp, nsi):
        match_info['reg0'] = (nsp << 8) | nsi
        # on header-matching there's no in_port
        match_info.pop('in_port', None)

    def _setup_local_switch_flows_on_int_br(self, flowrule,
                                            flow_classifier_list, actions,
                                            add_flow=True, match_inport=True):
        inport_match = {}
        priority = PC_DEF_PRI
        # no pp_corr means that classification will not be based on encap
        pp_corr = flowrule.get('pp_corr')
        node_type = flowrule['node_type']
        branch_info = flowrule.get('branch_info')
        on_add = None
        flow_count = 1
        if branch_info and node_type == constants.SRC_NODE:
            # for branching, we need as many flows (per flow info) as branches
            # because we can't AND-match the same field in a single flow
            flow_count = len(branch_info.get('matches'))
            on_add = branch_info.get('on_add')

        if match_inport is True:
            egress_port = self.br_int.get_vif_port_by_id(flowrule['egress'])
            if egress_port:
                inport_match = {'in_port': egress_port.ofport}
                priority = PC_INGRESS_PRI

        for flow_info in self._get_flow_infos_from_flow_classifier_list(
                flow_classifier_list, flowrule):
            match_info = dict(inport_match)
            match_info.update(flow_info)
            if node_type == constants.SF_NODE:
                if pp_corr:
                    match_info = {'in_port': match_info['in_port']}
                    if pp_corr == 'mpls':
                        match_info = self._build_classification_match_sfc_mpls(
                            flowrule, match_info)
                    elif pp_corr == 'nsh':
                        match_info = self._build_classification_match_sfc_nsh(
                            flowrule, match_info)

            for i in range(flow_count):
                if branch_info:
                    # for Service Graphs (branching):
                    nsp = branch_info['matches'][i][0]
                    nsi = branch_info['matches'][i][1]
                if add_flow:
                    if on_add:
                        self._match_by_header(match_info, nsp, nsi)
                    self.br_int.add_flow(table=ovs_consts.LOCAL_SWITCHING,
                                         priority=priority,
                                         actions=actions,
                                         **match_info)
                else:
                    if on_add is False:
                        self._match_by_header(match_info, nsp, nsi)
                    self.br_int.delete_flows(table=ovs_consts.LOCAL_SWITCHING,
                                             priority=priority,
                                             strict=True,
                                             **match_info)

    def _setup_egress_flow_rules(self, flowrule, match_inport=True):
        group_id = flowrule.get('next_group_id', None)
        next_hops = flowrule.get('next_hops', None)
        pc_corr = flowrule.get('pc_corr', 'mpls')
        pp_corr = flowrule.get('pp_corr', None)
        node_type = flowrule.get('node_type')
        next_hop_tap_enabled = None
        # if the group is not none, install the egress rule for this SF
        if group_id and next_hops:
            # 1st, install br-int flow rule on table ACROSS_SUBNET_TABLE
            # and group table
            buckets = []
            vlan = self._get_vlan_by_port(flowrule['egress'])

            if isinstance(next(iter(next_hops)), dict):
                next_hop_tap_enabled = next_hops[0].get('tap_enabled')

            for item in next_hops:
                # all next hops share same pp_corr, enforced by higher layers
                pp_corr_nh = item.get('pp_corr', None)
                if flowrule['fwd_path']:
                    bucket = (
                        'bucket=weight=%d, mod_dl_dst:%s, resubmit(,%d)' % (
                            item['weight'],
                            item['in_mac_address'],
                            ACROSS_SUBNET_TABLE))
                else:
                    bucket = (
                        'bucket=weight=%d, mod_dl_dst:%s, resubmit(,%d)' % (
                            item['weight'],
                            item['mac_address'],
                            ACROSS_SUBNET_TABLE))
                buckets.append(bucket)
                subnet_actions_list = []

                across_flow = "mod_vlan_vid:%d," % vlan
                # the classic encapsulation of packets in ACROSS_SUBNET_TABLE
                # is kept unchanged for the same scenarios, i.e. when the next
                # hops don't support encapsulation and neither the current one.
                if not pp_corr and pp_corr_nh is None:
                    if pc_corr == 'mpls':
                        push_encap = self._build_push_mpls(flowrule['nsp'],
                                                           flowrule['nsi'])
                    elif pc_corr == 'nsh':
                        push_encap = self._build_push_nsh(flowrule['nsp'],
                                                          flowrule['nsi'])
                    across_flow = push_encap + across_flow

                subnet_actions_list.append(across_flow)

                if item['local_endpoint'] == self.local_ip:
                    subnet_actions = 'resubmit(,%d)' % INGRESS_TABLE
                else:
                    # same subnet with next hop
                    subnet_actions = 'output:%s' % self.patch_tun_ofport
                subnet_actions_list.append(subnet_actions)

                eth_type = constants.ETH_TYPE_IP
                if pp_corr == 'mpls' or pp_corr_nh == 'mpls':
                    eth_type = constants.ETH_TYPE_MPLS
                elif pp_corr == 'nsh' or pp_corr_nh == 'nsh':
                    eth_type = constants.ETH_TYPE_NSH

                if item.get('tap_enabled'):
                    self._add_tap_classification_flows(flowrule,
                                                       item,
                                                       subnet_actions_list)
                else:
                    self._configure_across_subnet_flow(flowrule,
                                                       item,
                                                       subnet_actions_list,
                                                       eth_type)

            if not next_hop_tap_enabled:
                self._add_group_table(buckets, flowrule, group_id)

            # 2nd, install br-int flow rule on table 0 for egress traffic
            enc_actions = ""
            # we only encapsulate on table 0 if we know the next hops will
            # support that encapsulation but the current hop doesn't already.
            if not pp_corr and pp_corr_nh:
                if pc_corr == 'mpls':
                    enc_actions = self._build_push_mpls(flowrule['nsp'],
                                                        flowrule['nsi'])
                elif pc_corr == 'nsh':
                    enc_actions = self._build_push_nsh(flowrule['nsp'],
                                                       flowrule['nsi'])
            if flowrule['fwd_path']:
                enc_actions += "group:%d" % group_id
            else:
                rev_group_id = group_id + REVERSE_GROUP_NUMBER_OFFSET
                enc_actions += "group:%d" % rev_group_id

            enc_actions = self._update_enc_actions(enc_actions, flowrule,
                                                   next_hop_tap_enabled)
            # to uninstall the removed flow classifiers
            self._setup_local_switch_flows_on_int_br(
                flowrule,
                flowrule['del_fcs'],
                None,
                add_flow=False,
                match_inport=match_inport)
            # to install the added flow classifiers
            self._setup_local_switch_flows_on_int_br(
                flowrule,
                flowrule['add_fcs'],
                enc_actions,
                add_flow=True,
                match_inport=match_inport)
        else:
            end_of_chain_actions = 'normal'
            # at the end of the chain, the header must be removed (if used)
            if (node_type != constants.SRC_NODE) and pp_corr:
                branch_point = flowrule.get('branch_point')
                if branch_point:
                    nsp = flowrule['nsp']
                    nsi = flowrule['nsi']
                    sfpi = (nsp << 8) | nsi
                    if pc_corr == 'mpls':
                        end_of_chain_actions = (
                            'load:%s->NXM_NX_REG0[],'
                            'pop_mpls:0x%04x,resubmit(,0)' % (
                                hex(sfpi), constants.ETH_TYPE_IP))
                    elif pc_corr == 'nsh':
                        end_of_chain_actions = (
                            "load:%s->NXM_NX_REG0[],"
                            "decap(),decap(),resubmit(,0)" % (
                                hex(sfpi)))
                else:
                    if pc_corr == 'mpls':
                        end_of_chain_actions = ("pop_mpls:0x%04x,%s" % (
                                                constants.ETH_TYPE_IP,
                                                end_of_chain_actions))
                    elif pc_corr == 'nsh':
                        end_of_chain_actions = ("decap(),decap(),%s" % (
                                                end_of_chain_actions))

            if flowrule.get('tap_enabled'):
                end_of_chain_actions += RESUBMIT_TAP_TABLE

            # to uninstall the removed flow classifiers
            if 'del_fcs' in flowrule:
                self._setup_local_switch_flows_on_int_br(
                    flowrule,
                    flowrule['del_fcs'],
                    None,
                    add_flow=False,
                    match_inport=True)

            if 'add_fcs' in flowrule:
                # to install the added flow classifiers
                self._setup_local_switch_flows_on_int_br(
                    flowrule,
                    flowrule['add_fcs'],
                    actions=end_of_chain_actions,
                    add_flow=True,
                    match_inport=True)

    def _get_vlan_by_port(self, port_id):
        try:
            net_uuid = self.vlan_manager.get_net_uuid(port_id)
            return self.vlan_manager.get(net_uuid).vlan
        except (vlanmanager.VifIdNotFound, vlanmanager.MappingNotFound):
            return None

    def _setup_ingress_flow_rules(self, flowrule):
        vif_port = self.br_int.get_vif_port_by_id(flowrule['ingress'])
        if vif_port:
            vlan = self._get_vlan_by_port(flowrule['ingress'])
            pc_corr = flowrule['pc_corr']
            pp_corr = flowrule['pp_corr']

            # install br-int flow rule on table 0 for ingress traffic
            # install an SFC Proxy if the port pair doesn't support the
            # SFC encapsulation (pc_corr) specified in the chain
            if pc_corr == 'mpls':
                if flowrule.get('tap_enabled'):
                    return self._add_tap_ingress_flow(flowrule, vif_port, vlan)
                if pp_corr is None:
                    match_field = self._build_proxy_sfc_mpls(flowrule,
                                                             vif_port, vlan)
                elif pp_corr == 'mpls':
                    match_field = self._build_forward_sfc_mpls(flowrule,
                                                               vif_port, vlan)
            elif pc_corr == 'nsh':
                if pp_corr is None:
                    match_field = self._build_proxy_sfc_nsh(flowrule,
                                                            vif_port, vlan)
                elif pp_corr == 'nsh':
                    match_field = self._build_forward_sfc_nsh(flowrule,
                                                              vif_port, vlan)
            self.br_int.add_flow(**match_field)

    def _build_classification_match_sfc_mpls(self, flowrule, match_info):
        match_info['eth_type'] = constants.ETH_TYPE_MPLS
        match_info['mpls_label'] = flowrule['nsp'] << 8 | flowrule['nsi']
        return match_info

    def _build_classification_match_sfc_nsh(self, flowrule, match_info):
        match_info['eth_type'] = constants.ETH_TYPE_NSH
        match_info['nsh_mdtype'] = 1
        match_info['nsh_spi'] = flowrule['nsp']
        match_info['nsh_si'] = flowrule['nsi']
        return match_info

    def _build_push_mpls(self, nsp, nsi):
        return (
            "push_mpls:0x%04x,"
            "set_mpls_label:%d,"
            "set_mpls_ttl:%d," %
            (constants.ETH_TYPE_MPLS, nsp << 8 | nsi, nsi))

    def _build_push_nsh(self, nsp, nsi):
        return (
            "encap(nsh,prop(class=nsh,type=md_type,val=1)),"
            "set_field:%s->nsh_spi,set_field:%s->nsh_si,"
            "encap(ethernet)," %
            (hex(nsp), hex(nsi)))

    def _build_ingress_common_match_field(self, vif_port, vlan):
        return dict(
            table=INGRESS_TABLE,
            priority=1,
            dl_dst=vif_port.vif_mac,
            dl_vlan=vlan)

    def _build_ingress_match_field_sfc_mpls(self, flowrule, vif_port, vlan):
        match_field = self._build_ingress_common_match_field(vif_port, vlan)
        match_field['eth_type'] = constants.ETH_TYPE_MPLS
        match_field['mpls_label'] = flowrule['nsp'] << 8 | flowrule['nsi'] + 1
        return match_field

    def _build_ingress_match_field_sfc_nsh(self, flowrule, vif_port, vlan):
        match_field = self._build_ingress_common_match_field(vif_port, vlan)
        match_field['eth_type'] = constants.ETH_TYPE_NSH
        match_field['nsh_mdtype'] = 1
        match_field['nsh_spi'] = flowrule['nsp']
        match_field['nsh_si'] = flowrule['nsi'] + 1
        return match_field

    def _build_proxy_sfc_mpls(self, flowrule, vif_port, vlan):
        match_field = self._build_ingress_match_field_sfc_mpls(
            flowrule, vif_port, vlan)
        actions = ("strip_vlan, pop_mpls:0x%04x,"
                   "output:%s" % (constants.ETH_TYPE_IP, vif_port.ofport))
        match_field['actions'] = actions
        return match_field

    def _build_proxy_sfc_nsh(self, flowrule, vif_port, vlan):
        match_field = self._build_ingress_match_field_sfc_nsh(
            flowrule, vif_port, vlan)
        actions = ("strip_vlan,move:NXM_OF_ETH_DST->OXM_OF_PKT_REG0[0..47],"
                   "decap(),decap(),"
                   "move:OXM_OF_PKT_REG0[0..47]->NXM_OF_ETH_DST,output:%s"
                   "" % vif_port.ofport)
        match_field['actions'] = actions
        return match_field

    def _build_forward_sfc_mpls(self, flowrule, vif_port, vlan):
        match_field = self._build_ingress_match_field_sfc_mpls(
            flowrule, vif_port, vlan)
        actions = ("strip_vlan, output:%s" % vif_port.ofport)
        match_field['actions'] = actions
        return match_field

    def _build_forward_sfc_nsh(self, flowrule, vif_port, vlan):
        match_field = self._build_ingress_match_field_sfc_nsh(
            flowrule, vif_port, vlan)
        actions = ("strip_vlan, output:%s" % vif_port.ofport)
        match_field['actions'] = actions
        return match_field

    def _delete_flows_mpls(self, flowrule, vif_port):
        if flowrule.get('tap_enabled'):
            self.br_int.delete_flows(
                table=INGRESS_TABLE,
                eth_type=constants.ETH_TYPE_MPLS,
                dl_src=flowrule['mac_address'],
                mpls_label=flowrule['nsp'] << 8 | flowrule['nsi']
            )
        else:
            self.br_int.delete_flows(
                table=INGRESS_TABLE,
                eth_type=constants.ETH_TYPE_MPLS,
                dl_dst=vif_port.vif_mac,
                mpls_label=flowrule['nsp'] << 8 | flowrule['nsi'] + 1
            )

    def _add_group_table(self, buckets, flowrule, group_id):
        group_content = self.br_int.dump_group_for_id(group_id)
        buckets = ','.join(buckets)
        if flowrule['fwd_path']:
            if group_content.find('group_id=%d' % group_id) == -1:
                self.br_int.add_group(group_id=group_id,
                                      type='select',
                                      buckets=buckets)
            else:
                self.br_int.mod_group(group_id=group_id,
                                      type='select',
                                      buckets=buckets)
        else:
            # set different id for rev_group
            rev_group_id = group_id + REVERSE_GROUP_NUMBER_OFFSET
            if group_content.find('group_id=%d' % (rev_group_id)) == -1:
                self.br_int.add_group(group_id=rev_group_id,
                                      type='select',
                                      buckets=buckets)
            else:
                self.br_int.mod_group(group_id=rev_group_id,
                                      type='select',
                                      buckets=buckets)

    def _configure_across_subnet_flow(self, flowrule, item,
                                      subnet_actions_list, eth_type):
        if flowrule['fwd_path']:
            self.br_int.add_flow(
                table=ACROSS_SUBNET_TABLE,
                priority=0,
                dl_dst=item['in_mac_address'],
                eth_type=eth_type,
                actions="%s" % ','.join(subnet_actions_list))
        else:
            self.br_int.add_flow(
                table=ACROSS_SUBNET_TABLE,
                priority=0,
                dl_dst=item['mac_address'],
                eth_type=eth_type,
                actions="%s" % ','.join(subnet_actions_list))

    def _add_tap_classification_flows(self, flowrule, item,
                                      subnet_actions_list):
        egress_port = self.br_int.get_vif_port_by_id(flowrule['egress'])
        vlan = self._get_vlan_by_port(flowrule['egress'])
        if not egress_port:
            return
        in_port = egress_port.ofport
        vif_mac = egress_port.vif_mac

        tap_action = ""
        if flowrule['pc_corr'] == 'mpls':
            tap_action += self._build_push_mpls(item['nsp'], item['nsi'])
        tap_action += "mod_vlan_vid:%d," % vlan
        subnet_actions_list[0] = tap_action
        ovs_rule = dict()

        self._get_eth_type(flowrule, item, ovs_rule)
        ovs_rule.update(table=TAP_CLASSIFIER_TABLE,
                        priority=0,
                        in_port=in_port,
                        dl_src=vif_mac,
                        actions="%s" % ''.join(subnet_actions_list)
                        )
        self.br_int.add_flow(**ovs_rule)
        if item['local_endpoint'] != self.local_ip:
            self._configure_tunnel_bridge_flows(flowrule, item, vif_mac)

    def _get_eth_type(self, flowrule, item, ovs_rule):
        # eth_type is decided based on current node's pp_corr and next node of
        # Tap node's pp_corr
        if flowrule['pp_corr'] == item.get('pp_corr_tap_nh') is None:
            ovs_rule.update(eth_type=constants.ETH_TYPE_IP)
        elif flowrule['pp_corr'] and not item.get('pp_corr_tap_nh'):
            ovs_rule.update(eth_type=constants.ETH_TYPE_IP)
        elif not flowrule['pp_corr'] and item.get('pp_corr_tap_nh'):
            if item['pp_corr_tap_nh'] == 'mpls':
                eth_type = constants.ETH_TYPE_MPLS
                mpls_label = flowrule['nsp'] << 8 | flowrule['nsi']
                ovs_rule.update(mpls_label=mpls_label,
                                eth_type=eth_type)
        else:
            if flowrule['pp_corr'] == 'mpls' or item.get(
                    'pp_corr_tap_nh') == 'mpls':
                ovs_rule.update(eth_type=constants.ETH_TYPE_MPLS)

    def _configure_tunnel_bridge_flows(self, flowrule, item, vif_mac):
        local_tunnel_ports = [port for port in
                              self.br_tun.get_bridge_ports()
                              if port != self.patch_int_ofport]
        match_info = {'in_port': self.patch_int_ofport,
                      'dl_src': vif_mac}
        if flowrule['pc_corr'] == 'mpls':
            self._build_classification_match_sfc_mpls(item, match_info)
        self.br_tun.add_flow(
            table=0,
            priority=30,
            actions="resubmit(,%s)" % TAP_TUNNEL_OUTPUT_TABLE,
            **match_info
        )
        output_actions = "strip_vlan,load:%s->NXM_NX_TUN_ID[]" % (
            hex(flowrule['segment_id']))
        for port in local_tunnel_ports:
            output_actions += (",output:%d" % port)
        self.br_tun.add_flow(
            table=TAP_TUNNEL_OUTPUT_TABLE,
            priority=0,
            actions=output_actions,
            **match_info
        )

    def _build_buckets(self, buckets, flowrule, item):
        if item.get('tap_enabled'):
            # Tap PPG doesn't use bucket as of now.
            return
        if flowrule['fwd_path']:
            bucket = (
                'bucket=weight=%d, mod_dl_dst:%s, resubmit(,%d)' % (
                    item['weight'],
                    item['in_mac_address'],
                    ACROSS_SUBNET_TABLE))
        else:
            bucket = (
                'bucket=weight=%d, mod_dl_dst:%s, resubmit(,%d)' % (
                    item['weight'],
                    item['mac_address'],
                    ACROSS_SUBNET_TABLE))
        buckets.append(bucket)

    def _update_enc_actions(self, enc_actions, flow_rule,
                            next_hop_tap_enabled):
        # Add resubmit action to send to TAP table.
        if next_hop_tap_enabled:
            pp_corr = flow_rule['pp_corr']
            pp_corr_tap_nh = flow_rule['next_hops'][0].get('pp_corr_tap_nh')
            tap_nh_node_type = flow_rule['next_hops'][0].get(
                'tap_nh_node_type', constants.DST_NODE)
            group_action = enc_actions.split(',')[-1]
            enc_actions = ""
            if tap_nh_node_type == constants.SF_NODE:
                if not pp_corr and pp_corr_tap_nh:
                    if flow_rule.get('pc_corr', 'mpls') == 'mpls':
                        mpls_act = self._build_push_mpls(flow_rule['nsp'],
                                                         flow_rule['nsi'])
                        enc_actions += mpls_act
                    # enc_actions += group_action
                enc_actions += group_action
            else:
                if flow_rule['pc_corr'] == 'mpls':
                    # For DST Node
                    if flow_rule['pp_corr']:
                        enc_actions = ('pop_mpls:0x%04x,%s' % (
                            constants.ETH_TYPE_IP, NORMAL_ACTION))
                    else:
                        enc_actions += NORMAL_ACTION
            return enc_actions + RESUBMIT_TAP_TABLE
        elif flow_rule.get('tap_enabled'):
            return enc_actions + RESUBMIT_TAP_TABLE
        return enc_actions

    def _delete_across_subnet_table_flows(self, flowrule):
        if not flowrule['next_hops']:
            return
        tap_enabled = flowrule['next_hops'][0].get('tap_enabled', False)
        if tap_enabled:
            egress_port = self.br_int.get_vif_port_by_id(flowrule['egress'])
            for item in flowrule['next_hops']:
                if flowrule['fwd_path']:
                    self.br_int.delete_flows(
                        table=TAP_CLASSIFIER_TABLE,
                        dl_src=egress_port.vif_mac)
                else:
                    self.br_int.delete_flows(
                        table=TAP_CLASSIFIER_TABLE,
                        dl_src=egress_port.vif_mac)

                if item['local_endpoint'] != self.local_ip:
                    self._delete_tunnel_bridge_flows(flowrule,
                                                     egress_port.vif_mac)
        else:
            for item in flowrule['next_hops']:
                if flowrule['fwd_path']:
                    self.br_int.delete_flows(
                        table=ACROSS_SUBNET_TABLE,
                        dl_dst=item['in_mac_address'])
                else:
                    self.br_int.delete_flows(
                        table=ACROSS_SUBNET_TABLE,
                        dl_dst=item['mac_address'])

    def _add_tap_ingress_flow(self, flowrule, vif_port, vlan):
        match_field = self._build_tap_ingress_match_field_sfc_mpls(
            flowrule, vif_port, vlan)
        actions = ('strip_vlan, pop_mpls:0x%04x,output:%s'
                   % (constants.ETH_TYPE_MPLS,
                      vif_port.ofport))
        match_field['actions'] = actions
        match_field.pop('dl_dst', None)
        match_field.update(dl_src=flowrule['mac_address'])
        self.br_int.add_flow(**match_field)

    def _build_tap_ingress_match_field_sfc_mpls(self, flowrule, vif_port,
                                                vlan):
        match_field = self._build_ingress_common_match_field(vif_port, vlan)
        match_field['eth_type'] = constants.ETH_TYPE_MPLS
        match_field['mpls_label'] = flowrule['nsp'] << 8 | flowrule['nsi']
        return match_field

    def _delete_tunnel_bridge_flows(self, flowrule, src_mac):
        match_info = {'in_port': self.patch_int_ofport,
                      'dl_src': src_mac}
        # Use Tap 'nsi'
        flowrule_copy = flowrule.copy()
        flowrule_copy['nsi'], flowrule_copy['nsp'] = (
            flowrule['next_hops'][0]['nsi'], flowrule['next_hops'][0]['nsp'])
        self._build_classification_match_sfc_mpls(flowrule_copy, match_info)
        self.br_tun.delete_flows(table=0,
                                 **match_info)
        self.br_tun.delete_flows(table=TAP_TUNNEL_OUTPUT_TABLE,
                                 **match_info)

    def _clear_sfc_flow_on_tun_br(self):
        self.br_tun.delete_flows(table=0, eth_type=constants.ETH_TYPE_MPLS)
        self.br_tun.delete_flows(table=TAP_TUNNEL_OUTPUT_TABLE)

    def _delete_flows_nsh(self, flowrule, vif_port):
        self.br_int.delete_flows(
            table=INGRESS_TABLE,
            eth_type=constants.ETH_TYPE_NSH,
            dl_dst=vif_port.vif_mac,
            nsh_mdtype=1,
            nsh_spi=flowrule['nsp'],
            nsh_si=flowrule['nsi'] + 1
        )



# Copyright 2017 Futurewei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Exceptions used by SFC plugin and drivers."""

from neutron_lib import exceptions

from networking_sfc._i18n import _


class SfcDriverError(exceptions.NeutronException):
    """SFC driver call failed."""
    message = _("%(method)s failed.")


class SfcException(exceptions.NeutronException):
    """Base for SFC driver exceptions returned to user."""
    pass


class SfcBadRequest(exceptions.BadRequest, SfcException):
    """Base for SFC driver bad request exceptions returned to user."""
    message = _("%(message)s")


class SfcNoSubnetGateway(SfcDriverError):
    """No subnet gateway."""
    message = _("There is no %(type)s of ip prefix %(cidr)s.")


class SfcNoSuchSubnet(SfcDriverError):
    """No such subnet."""
    message = _("There is no %(type)s of %(cidr)s.")


class FlowClassifierInvalid(SfcDriverError):
    """Invalid flow classifier."""
    message = _("There is no %(type)s assigned.")



# Copyright 2017 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from mock import Mock
from stevedore.extension import Extension

from neutron.tests import base

from networking_sfc.services.sfc.common import exceptions as sfc_exc
from networking_sfc.services.sfc.driver_manager \
    import SfcDriverManager


class DriverManagerTestCase(base.BaseTestCase):
    def setUp(self):
        super(DriverManagerTestCase, self).setUp()

    def test_initialize_called(self):
        driver1 = Extension('mock_driver1', Mock(), None,
                            Mock(native_bulk_support=True))
        driver2 = Extension('mock_driver2', Mock(), None,
                            Mock(native_bulk_support=True))
        manager = SfcDriverManager.make_test_instance([driver1, driver2])
        manager.initialize()
        driver1.obj.initialize.assert_called_once_with()
        driver2.obj.initialize.assert_called_once_with()

    def _test_method_called(self, method_name):
        driver1 = Extension('mock_driver1', Mock(), None,
                            Mock(native_bulk_support=True))
        driver2 = Extension('mock_driver2', Mock(), None,
                            Mock(native_bulk_support=True))
        manager = SfcDriverManager.make_test_instance([driver1, driver2])
        mocked_context = Mock()
        getattr(manager, method_name)(mocked_context)
        getattr(driver1.obj, method_name).assert_called_once_with(
            mocked_context)
        getattr(driver2.obj, method_name).assert_called_once_with(
            mocked_context)

    def _test_method_exception(self, method_name,
                               expected_exc=sfc_exc.SfcDriverError):
        driver = Extension('mock_driver', Mock(), None,
                           Mock(native_bulk_support=True))
        mock_method = Mock(side_effect=sfc_exc.SfcException)
        setattr(driver.obj, method_name, mock_method)
        manager = SfcDriverManager.make_test_instance([driver])
        mocked_context = Mock()
        self.assertRaises(expected_exc,
                          getattr(manager, method_name),
                          mocked_context)

    def test_create_port_chain_precommit_called(self):
        self._test_method_called("create_port_chain_precommit")

    def test_create_port_chain_precommit_exception(self):
        self._test_method_exception("create_port_chain_precommit",
                                    sfc_exc.SfcException)

    def test_create_port_chain_postcommit_called(self):
        self._test_method_called("create_port_chain_postcommit")

    def test_create_port_chain_postcommit_exception(self):
        self._test_method_exception("create_port_chain_postcommit")

    def test_update_port_chain_precommit_called(self):
        self._test_method_called("update_port_chain_precommit")

    def test_update_port_chain_precommit_exception(self):
        self._test_method_exception("update_port_chain_precommit")

    def test_update_port_chain_postcommit_called(self):
        self._test_method_called("update_port_chain_postcommit")

    def test_update_port_chain_postcommit_exception(self):
        self._test_method_exception("update_port_chain_postcommit")

    def test_delete_port_chain_called(self):
        self._test_method_called("delete_port_chain")

    def test_delete_port_chain_exception(self):
        self._test_method_exception("delete_port_chain")

    def test_delete_port_chain_precommit_called(self):
        self._test_method_called("delete_port_chain_precommit")

    def test_delete_port_chain_precommit_exception(self):
        self._test_method_exception("delete_port_chain_precommit")

    def test_delete_port_chain_postcommit_called(self):
        self._test_method_called("delete_port_chain_postcommit")

    def test_delete_port_chain_postcommit_exception(self):
        self._test_method_exception("delete_port_chain_postcommit")

    def test_create_port_pair_group_precommit_called(self):
        self._test_method_called("create_port_pair_group_precommit")

    def test_create_port_pair_group_precommit_exception(self):
        self._test_method_exception("create_port_pair_group_precommit")

    def test_create_port_pair_group_postcommit_called(self):
        self._test_method_called("create_port_pair_group_postcommit")

    def test_create_port_pair_group_postcommit_exception(self):
        self._test_method_exception("create_port_pair_group_postcommit")

    def test_update_port_pair_group_precommit_called(self):
        self._test_method_called("update_port_pair_group_precommit")

    def test_update_port_pair_group_precommit_exception(self):
        self._test_method_exception("update_port_pair_group_precommit")

    def test_update_port_pair_group_postcommit_called(self):
        self._test_method_called("update_port_pair_group_postcommit")

    def test_update_port_pair_group_postcommit_exception(self):
        self._test_method_exception("update_port_pair_group_postcommit")

    def test_delete_port_pair_group_called(self):
        self._test_method_called("delete_port_pair_group")

    def test_delete_port_pair_group_exception(self):
        self._test_method_exception("delete_port_pair_group")

    def test_delete_port_pair_group_precommit_called(self):
        self._test_method_called("delete_port_pair_group_precommit")

    def test_delete_port_pair_group_precommit_exception(self):
        self._test_method_exception("delete_port_pair_group_precommit")

    def test_delete_port_pair_group_postcommit_called(self):
        self._test_method_called("delete_port_pair_group_postcommit")

    def test_delete_port_pair_group_postcommit_exception(self):
        self._test_method_exception("delete_port_pair_group_postcommit")

    def test_create_port_pair_precommit_called(self):
        self._test_method_called("create_port_pair_precommit")

    def test_create_port_pair_precommit_exception(self):
        self._test_method_exception("create_port_pair_precommit")

    def test_create_port_pair_postcommit_called(self):
        self._test_method_called("create_port_pair_postcommit")

    def test_create_port_pair_postcommit_exception(self):
        self._test_method_exception("create_port_pair_postcommit")

    def test_update_port_pair_precommit_called(self):
        self._test_method_called("update_port_pair_precommit")

    def test_update_port_pair_precommit_exception(self):
        self._test_method_exception("update_port_pair_precommit")

    def test_update_port_pair_postcommit_called(self):
        self._test_method_called("update_port_pair_postcommit")

    def test_update_port_pair_postcommit_exception(self):
        self._test_method_exception("update_port_pair_postcommit")

    def test_delete_port_pair_called(self):
        self._test_method_called("delete_port_pair")

    def test_delete_port_pair_exception(self):
        self._test_method_exception("delete_port_pair")

    def test_delete_port_pair_precommit_called(self):
        self._test_method_called("delete_port_pair_precommit")

    def test_delete_port_pair_precommit_exception(self):
        self._test_method_exception("delete_port_pair_precommit")

    def test_delete_port_pair_postcommit_called(self):
        self._test_method_called("delete_port_pair_postcommit")

    def test_delete_port_pair_postcommit_exception(self):
        self._test_method_exception("delete_port_pair_postcommit")

    def test_create_service_graph_precommit_called(self):
        self._test_method_called("create_service_graph_precommit")

    def test_create_service_graph_precommit_exception(self):
        self._test_method_exception("create_service_graph_precommit")

    def test_create_service_graph_postcommit_called(self):
        self._test_method_called("create_service_graph_postcommit")

    def test_create_service_graph_postcommit_exception(self):
        self._test_method_exception("create_service_graph_postcommit")

    def test_update_service_graph_precommit_called(self):
        self._test_method_called("update_service_graph_precommit")

    def test_update_service_graph_precommit_exception(self):
        self._test_method_exception("update_service_graph_precommit")

    def test_update_service_graph_postcommit_called(self):
        self._test_method_called("update_service_graph_postcommit")

    def test_update_service_graph_postcommit_exception(self):
        self._test_method_exception("update_service_graph_postcommit")

    def test_delete_service_graph_precommit_called(self):
        self._test_method_called("delete_service_graph_precommit")

    def test_delete_service_graph_precommit_exception(self):
        self._test_method_exception("delete_service_graph_precommit")

    def test_delete_service_graph_postcommit_called(self):
        self._test_method_called("delete_service_graph_postcommit")

    def test_delete_service_graph_postcommit_exception(self):
        self._test_method_exception("delete_service_graph_postcommit")



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class FlowClassifierDriverBaseLegacy(object):
    """Flow Classifier Driver Base Class for legacy driver interface"""

    @abc.abstractmethod
    def create_flow_classifier(self, context):
        pass

    @abc.abstractmethod
    def update_flow_classifier(self, context):
        pass


@six.add_metaclass(abc.ABCMeta)
class FlowClassifierDriverBase(FlowClassifierDriverBaseLegacy):
    """Flow Classifier Driver Base Class."""

    @abc.abstractmethod
    def create_flow_classifier_precommit(self, context):
        pass

    def create_flow_classifier_postcommit(self, context):
        self.create_flow_classifier(context)

    @abc.abstractmethod
    def delete_flow_classifier(self, context):
        pass

    def delete_flow_classifier_precommit(self, context):
        pass

    def delete_flow_classifier_postcommit(self, context):
        pass

    def update_flow_classifier_precommit(self, context):
        pass

    def update_flow_classifier_postcommit(self, context):
        self.update_flow_classifier(context)



# Copyright 2015 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys
from unittest import mock

from neutronclient import shell
from neutronclient.tests.unit import test_cli20

from networking_sfc.cli import flow_classifier as fc

from oslo_utils import uuidutils

source_port_UUID = uuidutils.generate_uuid()
destination_port_UUID = uuidutils.generate_uuid()


class CLITestV20FCExtensionJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        super(CLITestV20FCExtensionJSON, self).setUp()
        self._mock_extension_loading()
        self.register_non_admin_status_resource('flow_classifier')

    def _create_patch(self, name, func=None):
        patcher = mock.patch(name)
        thing = patcher.start()
        self.addCleanup(patcher.stop)
        return thing

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        flow_classifier = self._create_patch(ext_pkg +
                                             '._discover_via_entry_points')
        flow_classifier.return_value = [("flow_classifier", fc)]
        return flow_classifier

    def test_ext_cmd_loaded(self):
        neutron_shell = shell.NeutronShell('2.0')
        ext_cmd = {'flow-classifier-list': fc.FlowClassifierList,
                   'flow-classifier-create': fc.FlowClassifierCreate,
                   'flow-classifier-update': fc.FlowClassifierUpdate,
                   'flow-classifier-delete': fc.FlowClassifierDelete,
                   'flow-classifier-show': fc.FlowClassifierShow}
        for cmd_name, cmd_class in ext_cmd.items():
            found = neutron_shell.command_manager.find_command([cmd_name])
            self.assertEqual(cmd_class, found[0])

    def test_create_flow_classifier_with_mandatory_params(self):
        """create flow-classifier: flow1."""
        resource = 'flow_classifier'
        cmd = fc.FlowClassifierCreate(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        name = 'flow1'
        ethertype = 'IPv4'
        args = [
            name,
            '--ethertype', ethertype,
        ]
        position_names = ['name', 'ethertype']
        position_values = [name, ethertype]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_create_flow_classifier_with_all_params(self):
        """create flow-classifier: flow1."""
        resource = 'flow_classifier'
        cmd = fc.FlowClassifierCreate(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        name = 'flow1'
        protocol_name = 'TCP'
        ethertype = 'IPv4'
        source_port = '0:65535'
        source_port_min = 0
        source_port_max = 65535
        destination_port = '1:65534'
        destination_port_min = 1
        destination_port_max = 65534
        source_ip = '192.168.1.0/24'
        destination_ip = '192.168.2.0/24'
        logical_source_port = '4a334cd4-fe9c-4fae-af4b-321c5e2eb051'
        logical_destination_port = '1278dcd4-459f-62ed-754b-87fc5e4a6751'
        description = 'my-desc'
        l7_param = "url=my_url"
        l7_param_expected = {"url": "my_url"}
        args = [name,
                '--protocol', protocol_name,
                '--ethertype', ethertype,
                '--source-port', source_port,
                '--destination-port', destination_port,
                '--source-ip-prefix', source_ip,
                '--destination-ip-prefix', destination_ip,
                '--logical-source-port', logical_source_port,
                '--logical-destination-port', logical_destination_port,
                '--description', description,
                '--l7-parameters', l7_param]
        position_names = ['name', 'protocol', 'ethertype',
                          'source_port_range_min', 'source_port_range_max',
                          'destination_port_range_min',
                          'destination_port_range_max',
                          'source_ip_prefix', 'destination_ip_prefix',
                          'logical_source_port', 'logical_destination_port',
                          'description', 'l7_parameters']
        position_values = [name, protocol_name, ethertype,
                           source_port_min, source_port_max,
                           destination_port_min, destination_port_max,
                           source_ip, destination_ip,
                           logical_source_port,
                           logical_destination_port, description,
                           l7_param_expected]
        self._test_create_resource(resource, cmd, name, myid, args,
                                   position_names, position_values)

    def test_list_flow_classifier(self):
        """List available flow-classifiers."""
        resources = "flow_classifiers"
        cmd = fc.FlowClassifierList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_list_flow_classifier_sort(self):
        """flow_classifier-list --sort-key name --sort-key id --sort-key asc

        --sort-key desc
        """
        resources = "flow_classifiers"
        cmd = fc.FlowClassifierList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"])

    def test_list_flow_classifier_limit(self):
        """flow-classifier-list -P."""
        resources = "flow_classifiers"
        cmd = fc.FlowClassifierList(test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, page_size=1000)

    def test_show_flow_classifier_id(self):
        """flow-classifier-show test_id."""
        resource = 'flow_classifier'
        cmd = fc.FlowClassifierShow(test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args, ['id'])

    def test_show_flow_classifier_id_name(self):
        """flow-classifier-show ."""
        resource = 'flow_classifier'
        cmd = fc.FlowClassifierShow(test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id,
                                 args, ['id', 'name'])

    def test_update_flow_classifier_description(self):
        """flow-classifier-update myid --description mydesc."""
        resource = 'flow_classifier'
        cmd = fc.FlowClassifierUpdate(test_cli20.MyApp(sys.stdout), None)
        myid = 'myid'
        args = [myid, '--description', 'flow_classifier1', '--description',
                'flow_classifier2']
        updatefields = {'description': 'flow_classifier2'}
        self._test_update_resource(resource, cmd, myid, args, updatefields)

    def test_update_flow_classifier_name(self):
        """flow-classifier-update myid --name myname."""
        resource = 'flow_classifier'
        cmd = fc.FlowClassifierUpdate(test_cli20.MyApp(sys.stdout), None)
        self._test_update_resource(resource, cmd, 'myid',
                                   ['myid', '--name', 'myname'],
                                   {'name': 'myname'})

    def test_delete_flow_classifer(self):
        """flow-classifier-delete my-id."""
        resource = 'flow_classifier'
        cmd = fc.FlowClassifierDelete(test_cli20.MyApp(sys.stdout), None)
        my_id = 'myid1'
        args = [my_id]
        self._test_delete_resource(resource, cmd, my_id, args)



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions

from neutron.tests import base

from networking_sfc.services.sfc.common import ovs_ext_lib


class GetPortMaskTestCase(base.BaseTestCase):
    def setUp(self):
        super(GetPortMaskTestCase, self).setUp()

    def tearDown(self):
        super(GetPortMaskTestCase, self).tearDown()

    def test_single_port(self):
        masks = ovs_ext_lib.get_port_mask(100, 100)
        self.assertEqual(['0x64/0xffff'], masks)

    def test_invalid_min_port(self):
        self.assertRaises(
            exceptions.InvalidInput,
            ovs_ext_lib.get_port_mask,
            0, 100
        )

    def test_invalid_max_port(self):
        self.assertRaises(
            exceptions.InvalidInput,
            ovs_ext_lib.get_port_mask,
            100, 65536
        )

    def test_invalid_port_range(self):
        self.assertRaises(
            exceptions.InvalidInput,
            ovs_ext_lib.get_port_mask,
            100, 99
        )

    def test_one_port_mask(self):
        masks = ovs_ext_lib.get_port_mask(100, 101)
        self.assertEqual(['0x64/0xfffe'], masks)
        masks = ovs_ext_lib.get_port_mask(100, 103)
        self.assertEqual(['0x64/0xfffc'], masks)
        masks = ovs_ext_lib.get_port_mask(32768, 65535)
        self.assertEqual(['0x8000/0x8000'], masks)

    def test_multi_port_masks(self):
        masks = ovs_ext_lib.get_port_mask(101, 102)
        self.assertEqual(['0x65/0xffff', '0x66/0xffff'], masks)
        masks = ovs_ext_lib.get_port_mask(101, 104)
        self.assertEqual(
            ['0x65/0xffff', '0x66/0xfffe', '0x68/0xffff'],
            masks
        )
        masks = ovs_ext_lib.get_port_mask(1, 65535)
        self.assertEqual(
            [
                '0x1/0xffff',
                '0x2/0xfffe',
                '0x4/0xfffc',
                '0x8/0xfff8',
                '0x10/0xfff0',
                '0x20/0xffe0',
                '0x40/0xffc0',
                '0x80/0xff80',
                '0x100/0xff00',
                '0x200/0xfe00',
                '0x400/0xfc00',
                '0x800/0xf800',
                '0x1000/0xf000',
                '0x2000/0xe000',
                '0x4000/0xc000',
                '0x8000/0x8000'
            ],
            masks
        )
        masks = ovs_ext_lib.get_port_mask(32767, 65535)
        self.assertEqual(
            ['0x7fff/0xffff', '0x8000/0x8000'],
            masks
        )
